def alphabets(type_nuc=True):
    """
    provide ambiguity code for sequences
    """

    nucleotide_alphabet = ['A', 'C', 'G', 'T']

    nucleotide_alphabet_full = [
        'A',
        'C',
        'G',
        'T',
        'N',
        'B',
        'D',
        'H',
        'V',
        'Y',
        'R',
        'W',
        'S',
        'K',
        'M',
    ]

    nucleotide_ambiguity_code = {
        'N': ['A', 'C', 'G', 'T'],  # any
        'B': ['C', 'G', 'T'],  # not A
        'D': ['A', 'G', 'T'],  # not C
        'H': ['A', 'C', 'T'],  # not G
        'V': ['A', 'C', 'G'],  # not T
        'Y': ['C', 'T'],  # pyrimidine
        'R': ['A', 'G'],  # purine
        'W': ['A', 'T'],  # weak
        'S': ['C', 'G'],  # strong
        'K': ['G', 'T'],  # keto
        'M': ['A', 'C'],
    }  # amino

    nucleotide_match_dict = {
        'N': '[ACGTNBDHVYRWSKM]',  # any
        'B': '[CGTNBDHVYRWSKM]',  # not A
        'D': '[AGTNBDHVYRWSKM]',  # not C
        'H': '[ACTNBDHVYRWSKM]',  # not G
        'V': '[ACGNBDHVYRWSKM]',  # not T
        'K': '[GTNBDHVYRWSK]',  # keto - not A,C,M
        'M': '[ACNBDHVYRWSM]',  # amino - not G,T,K
        'W': '[ATNBDHVYRWKM]',  # weak - not C,G,S
        'S': '[CGNBDHVYRSKM]',  # strong - not A,G,W
        'Y': '[CTNBDHVYWSKM]',  # pyrimidine - not A,G,R
        'R': '[AGNBDHVRWSKM]',  # purine - not C,T,Y
        'A': '[ANDHVRWM]',
        'C': '[CNBHVYSM]',
        'G': '[GNBDVRSK]',
        'T': '[TNBDHYWK]',
    }

    aminoacid_alphabet = [
        'A',
        'R',
        'N',
        'D',
        'C',
        'E',
        'Q',
        'G',
        'H',
        'I',
        'L',
        'K',
        'M',
        'F',
        'P',
        'S',
        'T',
        'W',
        'Y',
        'V',
        'U',
        'O',
        '*',
    ]

    aminoacid_alphabet_full = [
        'A',
        'R',
        'N',
        'D',
        'C',
        'E',
        'Q',
        'G',
        'H',
        'I',
        'L',
        'K',
        'M',
        'F',
        'P',
        'S',
        'T',
        'W',
        'Y',
        'V',
        'U',
        'O',
        '*',
        'J',
        'Z',
        'B',
        'X',
    ]

    aminoacid_ambiguity_code = {
        'J': ['I', 'L'],
        'Z': ['Q', 'E'],
        'B': ['N', 'D'],
        'X': [
            'A',
            'R',
            'N',
            'D',
            'C',
            'E',
            'Q',
            'G',
            'H',
            'I',
            'L',
            'K',
            'M',
            'F',
            'P',
            'S',
            'T',
            'W',
            'Y',
            'V',
            'U',
            'O',
            '*',
        ],
    }  # any

    aminoacid_match_dict = {
        'J': '[ILJ]',
        'Z': '[QEZ]',
        'B': '[NDB]',
        # "X": ".",
        'X': '[ARNDCEQGHILKMFPSTWYVUO*XBZJ]',
        'A': '[AX]',
        'R': '[RX]',
        'N': '[NXB]',
        'D': '[DXB]',
        'C': '[CX]',
        'E': '[EXZ]',
        'Q': '[QXZ]',
        'G': '[GX]',
        'H': '[HX]',
        'I': '[IXJ]',
        'L': '[LXJ]',
        'K': '[KX]',
        'M': '[MX]',
        'F': '[FX]',
        'P': '[PX]',
        'S': '[SX]',
        'T': '[TX]',
        'W': '[WX]',
        'Y': '[YX]',
        'V': '[VX]',
        'U': '[UX]',
        'O': '[OX]',
        '*': '[*X]',
    }

    # aa_only = {'E', 'F', 'I', 'J', 'L', 'O', 'Q', 'P', 'U', 'X', 'Z', '*'}
    # return nucleotide_alphabet, nucleotide_alphabet_full, nucleotide_ambiguity_code, aminoacid_alphabet, aminoacid_alphabet_full, aminoacid_ambiguity_code, aa_only

    if type_nuc:
        return (
            nucleotide_alphabet,
            nucleotide_alphabet_full,
            nucleotide_ambiguity_code,
            nucleotide_match_dict,
        )
    else:
        return (
            aminoacid_alphabet,
            aminoacid_alphabet_full,
            aminoacid_ambiguity_code,
            aminoacid_match_dict,
        )



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

from flexidot._version import __version__


def parse_args():
    parser = ArgumentParser(
        prog='flexidot',
        description='FlexiDot: Flexible dotplot generation tool',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    # Input/output arguments
    parser.add_argument(
        '-i',
        '--infiles',
        required=True,
        nargs='+',
        help='Input fasta files (fasta file name or space-separated file list.)',
    )
    parser.add_argument(
        '-o',
        '--output_prefix',
        default='flexidot_output',
        help='File prefix to be added to the generated filenames.',
    )
    parser.add_argument(
        '--outdir',
        default='.',
        help='Output directory. Default: current directory.',
    )

    # Collage arguments
    parser.add_argument(
        '-c',
        '--collage',
        action='store_true',
        default=False,
        help='Combine multiple dotplots in a collage.',
    )
    parser.add_argument(
        '--n_col',
        type=int,
        default=4,
        help='Number of columns per page (if collage is ON.',
    )
    parser.add_argument(
        '--n_row',
        type=int,
        default=5,
        help='Number of rows per page (if collage is ON).',
    )

    # File type
    parser.add_argument(
        '-f',
        '--filetype',
        choices=['png', 'pdf', 'svg'],
        default='png',
        help='Output file format: png, pdf, svg',
    )

    # Sorting
    parser.add_argument(
        '-s',
        '--sort',
        action='store_true',
        default=False,
        help='Sort sequences alphabetically by name.',
    )

    # Calculation parameters
    parser.add_argument(
        '-k',
        '--wordsize',
        type=int,
        default=10,
        help='Wordsize (kmer length) for dotplot comparison.',
    )
    parser.add_argument(
        '-m',
        '--mode',
        action='append',
        choices=['0', '1', '2'],
        help='Mode of FlexiDot dotplotting. 0 = self [default], 1 = paired, 2 = poly (matrix with all-against-all dotplots). Call -m multiple times to run multiple modes.',
    )
    parser.add_argument(
        '-t',
        '--type_seq',
        choices=['aa', 'nuc'],
        default='nuc',
        help='Biological sequence type: aa (amino acid) or nuc (nucleotide).',
    )
    parser.add_argument(
        '-w',
        '--wobble_conversion',
        action='store_true',
        default=False,
        help='Ambiguity handling for relaxed matching. Note: This may make kmer matching slower.',
    )
    parser.add_argument(
        '-S',
        '--substitution_count',
        type=int,
        default=0,
        help='Number of substitutions allowed per window.',
    )
    parser.add_argument(
        '--max_n',
        type=float,
        default=10,
        help='Maximum percentage of Ns allowed in a kmer window. Applies only if --wobble_conversion is set, else kmers with Ns are skipped. Default: 10',
    )
    parser.add_argument(
        '-r',
        '--norev',
        action='store_true',
        default=False,
        help='Do not calculate reverse complementary matches (only for nucleotide sequences.)',
    )
    parser.add_argument(
        '-O',
        '--only_vs_first_seq',
        action='store_true',
        default=False,
        help='Limit pairwise comparisons to the 1st sequence only (if plotting mode=1 paired.)',
    )

    # Graphic formatting
    parser.add_argument('-A', '--line_width', type=float, default=1, help='Line width')

    parser.add_argument('-B', '--line_col_for', default='black', help='Line color')

    parser.add_argument(
        '-C', '--line_col_rev', default='green', help='Reverse line color'
    )

    parser.add_argument(
        '-D',
        '--x_label_pos',
        choices=['top', 'bottom'],
        default='top',
        help="Position of the X-label. Default: 'top'",
    )

    parser.add_argument('-E', '--label_size', type=int, default=10, help='Font size')

    parser.add_argument(
        '-F',
        '--spacing',
        type=float,
        default=0.04,
        help='Spacing between dotplots (if plotting mode=2 polyplot).',
    )

    parser.add_argument(
        '-L',
        '--length_scaling',
        action='store_true',
        default=False,
        help='Scale plot size for pairwise comparison.',
    )

    parser.add_argument(
        '-M',
        '--mirror_y_axis',
        action='store_true',
        default=False,
        help='Flip y-axis (bottom-to-top or top-to-bottom)',
    )

    parser.add_argument('-P', '--plot_size', type=int, default=10, help='Plot size')

    parser.add_argument(
        '-R',
        '--representation',
        type=int,
        choices=[0, 1, 2],
        default=0,
        help='Region of plot to display. Only if plotting mode is 2: polyplot\n\
                                    0 = full [default]\n\
                                    1 = upper\n\
                                    2 = lower',
    )

    parser.add_argument(
        '-T',
        '--title_length',
        type=int,
        default=50,
        help='Limit title length for comparisons. Default: 50 characters',
    )

    # GFF shading
    parser.add_argument(
        '-g',
        '--gff',
        nargs='+',
        default=None,
        help='GFF3 files for markup in self-dotplots. Provide a space-delimited list of GFF files.',
    )
    parser.add_argument(
        '-G',
        '--gff_color_config',
        default=None,
        help='Config file for custom GFF shading.',
    )

    # longest common subsequence (LCS) shading
    parser.add_argument(
        '-x',
        '--lcs_shading',
        action='store_true',
        default=False,
        help='Shade subdotplot based on longest common subsequence (LCS).',
    )

    parser.add_argument(
        '-X',
        '--lcs_shading_num',
        type=int,
        default=5,
        help='Number of shading intervals.',
    )

    parser.add_argument(
        '-y',
        '--lcs_shading_ref',
        type=int,
        choices=[0, 1, 2],
        default=0,
        help='Reference for LCS shading.\n\
                        0 = maximal LCS length [default]\n\
                        1 = maximally possible length (length of shorter sequence in pairwise comparison) \n\
                        2 = given interval sizes - DNA [default 100 bp] or proteins [default 10 aa] - see -Y',
    )

    parser.add_argument(
        '-Y',
        '--lcs_shading_interval_len',
        type=int,
        default=50,
        help='Length of intervals for LCS shading (only if --lcs_shading_ref=2) [default for nucleotides = 50; default for amino acids = 10]',
    )

    parser.add_argument(
        '-z',
        '--lcs_shading_ori',
        type=int,
        choices=[0, 1, 2],
        default=0,
        help='Shade subdotplots based on LCS\n\
                            0 = forward [default]\n\
                            1 = reverse, or\n\
                            2 = both strands (forward shading above diagonal, reverse shading on diagonal and below; if using --user_matrix_file, best LCS is used below diagonal)',
    )

    # Custom matrix shading
    parser.add_argument(
        '-u', '--user_matrix_file', help='Matrix file for shading above diagonal.'
    )
    parser.add_argument(
        '-U',
        '--user_matrix_print',
        action='store_true',
        default=False,
        help='Display matrix entries above diagonal.',
    )

    # Other options
    parser.add_argument(
        '--loglevel',
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help="Set logging level. Default: 'INFO'",
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='%(prog)s {version}'.format(version=__version__),
    )
    parser.add_argument('--logfile', default=None, help='Name of log file')

    return parser.parse_args()



import logging
import os

from Bio import SeqIO
import matplotlib.colors as mcolors
import pylab as P


def read_seq(input_fasta, degap=False):
    """
    Read fasta sequences from (all) file(s)
    """

    # Check if file provided
    if input_fasta == [] or input_fasta == '':
        raise ValueError('No file names provided: %s' % input_fasta)

    # Initialize variables
    concat_created = False

    # Combine sequence files, if required
    if type(input_fasta) is list:
        # Concatenate fasta files
        if len(input_fasta) > 1:
            logging.info(
                'Concatenating sequences from multiple files: {}'.format(input_fasta)
            )
            input_fasta_combi = concatenate_files(input_fasta)
            concat_created = True
        else:
            input_fasta_combi = input_fasta[0]
    else:
        # Single fasta file
        input_fasta_combi = input_fasta

    # read sequences
    logging.info(f'Reading sequences from {input_fasta_combi}')
    try:
        # SeqIO.index returns a dictionary with the sequence id as key and the sequence as value
        seq_dict = SeqIO.index(input_fasta_combi, 'fasta')
    except ValueError as e:
        logging.error(
            f'ValueError: {e} - please check input files, e.g. for duplicate names!'
        )
        return {}, []
    except FileNotFoundError as e:
        logging.error(f'FileNotFoundError: {e} - please check if the file exists!')
        return {}, []
    except Exception as e:
        logging.error(f'Unexpected error: {e} - please check input files!')
        return {}, []

    for seq_id in seq_dict:
        if '-' in seq_dict[seq_id].seq:
            logging.warning('Gaps detected in sequence: %s' % seq_id)
            return read_seq(degap_fasta(input_fasta), degap=True)

    # Get sequence names for sorting
    sequences = []
    for item in SeqIO.parse(input_fasta_combi, 'fasta'):
        sequences.append(item.id)

    # If degap=True remove input file after processing
    if degap and type(input_fasta) is list:
        logging.info('Removing temp degapped input fasta files: {}'.format(input_fasta))
        for item in input_fasta:
            os.remove(item)
    elif degap:
        logging.info('Removing temp degapped input fasta file: {}'.format(input_fasta))
        os.remove(input_fasta)

    # If concatenation was required, remove combined file
    if concat_created:
        logging.info('Removing concatenated tempfile: {}'.format(input_fasta_combi))
        os.remove(input_fasta_combi)

    return seq_dict, sequences


def read_gff_color_config(gff_color_config_file=''):
    """
    define coloring options for gff-based color shading of self-dotplots
    """

    # default aestetics for annotation shading (e.g. if no user config file is provided)
    # dictionary with feature_type as key and tuple(color, transparency, zoom) as value
    gff_feat_colors = {
        'orf': ('#b41a31', 0.2, 0),
        'orf_rev': ('#ff773b', 0.3, 0),
        'gene': ('#b41a31', 0.2, 0),
        'cds': ('darkorange', 0.2, 0),
        'exon': ('orange', 0.2, 0),
        'intron': ('lightgrey', 0.2, 0),
        'utr': ('lightblue', 0.2, 0),
        'repeat_region': ('green', 0.3, 0),
        'repeat': ('green', 0.3, 0),
        'tandem_repeat': ('red', 0.3, 0),
        'transposable_element': ('blue', 0.3, 0),
        'ltr_retrotransposon': ('#cccccc', 0.5, 0),
        'ltr-retro': ('#cccccc', 0.5, 0),
        'long_terminal_repeat': ('#2dd0f0', 0.75, 2),
        'ltr': ('#2dd0f0', 0.75, 2),
        'pbs': ('purple', 0.75, 2),
        'ppt': ('#17805a', 0.5, 2),
        'target_site_duplication': ('red', 0.75, 2),
        'misc_feature': ('grey', 0.3, 0),
        'misc_feat': ('grey', 0.3, 0),
        'misc': ('grey', 0.3, 0),
        'others': ('grey', 0.5, 0),
    }
    if not gff_color_config_file or not os.path.exists(str(gff_color_config_file)):
        logging.info('No custom GFF color configuration found. Using defaults.')
        return gff_feat_colors

    logging.info('Updating GFF color configuration with custom specifications.')

    # Read custom gff_color_config_file
    in_file = open(gff_color_config_file, 'r')
    overwritten = set()
    for line in in_file:
        if len(line.strip().split('\t')) >= 4 and not line.startswith('#'):
            data = line.strip().split('\t')
            feat = data[0].lower()
            color = data[1].lower()

            # Check, if settings are valid
            if not mcolors.is_color_like(color):
                color = 'grey'
                text = 'Invalid color specified for %s: %s - default grey' % (
                    data[0],
                    data[1],
                )
                logging.info(text)
            try:
                alpha = float(data[2])
            except ValueError:
                alpha = 0.75
                text = 'Invalid alpha specified for %s: %s - default 0.75' % (
                    data[0],
                    data[2],
                )
                logging.info(text)
            try:
                zoom = float(data[3])
            except Exception:
                zoom = 0
                text = 'Invalid zoom specified for %s: %s - default 0' % (
                    data[0],
                    data[3],
                )
                logging.info(text)

            # Track changes of predefined settings
            if feat in list(gff_feat_colors.keys()):
                overwritten.add(data[0].lower())

            gff_feat_colors[feat] = (color, alpha, zoom)

    in_file.close()

    # Default coloring for unknown annotations
    if 'others' not in list(gff_feat_colors.keys()):
        gff_feat_colors['others'] = ('grey', 0.5, 0)

    # Print configuration
    text = '\n\nGFF color specification:\n%s\n' % (60 * '.')
    for item in sorted(gff_feat_colors.keys()):
        text += '%-30s\t%-10s\t%-5s\t%s\n' % (
            item,
            str(gff_feat_colors[item][0]),
            str(gff_feat_colors[item][1]),
            str(gff_feat_colors[item][2]),
        )
    logging.debug(text)

    # print overwritting feature type specifications
    if len(overwritten) != 0:
        text = '%d feature type specifications overwritten:' % len(overwritten)
        text += '\n\t' + ', '.join(overwritten) + '\n'
        logging.info(text)

    text = 'GFF color specification updated acc. to %s\n\t%s\n\n' % (
        gff_color_config_file,
        ', '.join(gff_feat_colors),
    )
    logging.info(text)

    return gff_feat_colors


def read_gffs(
    input_gff_files,
    color_dict=None,
    type_nuc=True,
    prefix='',
    filetype='png',
):
    """
    Create feature dictionary from input_gff
    sequence name as key and (feature type, start, stop) as value
    """
    if color_dict is None:
        color_dict = {'others': ('grey', 1, 0)}

    if type(input_gff_files) is not list:
        input_gff_files = [input_gff_files]

    # Create dictionary with seq_name as key and (type, start and stop) as value
    unknown_feats = set()
    used_feats = set()
    feat_dict = {}
    for input_gff in input_gff_files:
        text = '...reading ' + input_gff
        logging.info(text)

        in_file = open(input_gff, 'r')
        for line in in_file:
            if not line.startswith('#') and line.strip() != '':
                data = line.strip().split('\t')
                feat_type = data[2].lower()
                if data[6] == '-':
                    feat_type += '_rev'
                if feat_type.lower() not in list(color_dict.keys()):
                    if feat_type.lower().replace('_rev', '') in list(color_dict.keys()):
                        feat_type = feat_type.replace('_rev', '')
                    else:
                        unknown_feats.add(feat_type)
                        feat_type = 'others'
                used_feats.add(feat_type)
                if data[0] not in list(feat_dict.keys()):
                    feat_dict[data[0]] = [
                        (feat_type, int(data[3]), int(data[4]))
                    ]  # feature type, start, stop
                else:
                    feat_dict[data[0]].append(
                        (feat_type, int(data[3]), int(data[4]))
                    )  # feature type, start, stop

        text = '\nAnnotations for: %s\n' % ', '.join(list(feat_dict.keys())[:10])
        if len(list(feat_dict.keys())) > 10:
            text = text[:-1] + ', ...\n'
        logging.info(text)

        in_file.close()

    # print feature types without specific shading settings
    if len(unknown_feats) != 0:
        text = 'Missing shading specification for %d feature type(s):\n\t%s\n' % (
            len(unknown_feats),
            ', '.join(sorted(unknown_feats)),
        )
        logging.info(text)

    # create color legend
    colors, alphas = [], []
    for item in sorted(used_feats):
        colors.append(color_dict[item][0])
        alphas.append(color_dict[item][1])
    legend_figure(
        colors=colors,
        lcs_shading_num=len(used_feats),
        type_nuc=type_nuc,
        bins=sorted(used_feats),
        alphas=alphas,
        gff_legend=True,
        prefix=prefix,
        filetype=filetype,
    )

    # print settings
    text = 'GFF Feature Types: %s\nGFF Colors:        %s' % (
        ', '.join(sorted(used_feats)),
        ', '.join(sorted(colors)),
    )
    logging.info(text)

    return feat_dict


def read_matrix(matrix_file_name, delim='\t', symmetric=True, recursion=False):
    input_file = open(matrix_file_name, 'r')

    # read sequence names from first column
    names = []
    for line in input_file:
        if not line.startswith('#') and not line.startswith(delim) and delim in line:
            names.append(line.strip().split(delim)[0])
    logging.info(
        "Delimiter '%s': %d names - %s\n" % (delim, len(names), ', '.join(names))
    )

    # check if names were found - otherwise try another delimiter
    if names == [] and not recursion:
        if delim == '\t':
            new_delim = ','
        else:
            new_delim = '\t'
        logging.info(
            "\nMatrix file not containing data delimited by '%s' - trying to read matrix with delimiter '%s'"
            % (delim.replace('\t', '\\t'), new_delim)
        )
        info_dict = read_matrix(
            matrix_file_name,
            delim=new_delim,
            symmetric=symmetric,
            recursion=True,
        )
        return info_dict
    elif names == []:
        logging.info('Empty matrix file with alternative delimiter!')
        return info_dict
    input_file.close()

    input_file = open(matrix_file_name, 'r')
    # read matrix entries as values in dictionary with tuple(names) as key
    info_dict = {}
    contradictory_entries = []
    for line in input_file:
        if not line.startswith('#') and not line.startswith(delim) and delim in line:
            data = line.strip().split(delim)
            for idx in range(len(data[1:])):
                # print tuple(sorted([data[0], names[idx]])), data[idx+1]
                if symmetric:
                    key = tuple(sorted([names[idx], data[0]]))
                else:
                    key = tuple(names[idx], data[0])
                if key in list(info_dict.keys()):
                    if (
                        symmetric
                        and info_dict[key] != data[idx + 1]
                        and data[idx + 1] not in ['', '-']
                        and info_dict[key] not in ['', '-']
                    ):
                        contradictory_entries.append(key)
                info_dict[key] = data[idx + 1]
    input_file.close()

    if len(contradictory_entries) != 0:
        try:
            logging.info(
                '\nContradictory entries in matrix file %s:\n\t%s'
                % (matrix_file_name, ', '.join(contradictory_entries))
            )
        except Exception:
            log_txt = '\nContradictory entries in matrix file %s:\n\t' % (
                matrix_file_name
            )
            for item in contradictory_entries:
                log_txt += str(item).replace("'", '') + ', '
            log_txt = log_txt[:-2]
            logging.info(log_txt)
        logging.info('Using value from bottom left triangle!')

    logging.debug(f'Matrix information: {", ".join(names)}')

    return info_dict


def concatenate_files(file_list, combi_filename='temp_combined.fasta'):
    """
    concatenate content of all files in file_list into a combined file named combi_filename
    """
    out_file = open(combi_filename, 'w')
    text = ''
    for item in file_list:
        text += item + ' '
        # read in_file linewise and write to out_file
        in_file = open(item, 'r')
        for line in in_file:
            out_file.write(line.strip() + '\n')
        in_file.close()
    out_file.close()

    logging.debug(text)

    return combi_filename


def degap_fasta(input_fasta):
    """
    Remove gaps from fasta - Write new degapped fasta.
    """
    logging.info('Removing gaps from fasta files: %s' % input_fasta)
    # degap all sequence files
    output_fastas = []
    if type(input_fasta) is not list:
        input_fasta = list(input_fasta)
    for input_fas in input_fasta:
        output_fas = input_fas[: input_fas.rfind('.')] + '_degapped.fas'
        in_file = open(input_fas, 'r')
        out_file = open(output_fas, 'w')
        for line in in_file:
            if line.startswith('>'):
                out_file.write(line.strip() + '\n')
            else:
                out_file.write(line.strip().replace('-', '') + '\n')
        out_file.close()
        in_file.close()
        output_fastas.append(output_fas)
    return output_fastas


def legend_figure(
    colors,
    lcs_shading_num,
    type_nuc=True,
    unit='%',
    filetype='png',
    max_len=None,
    min_len=0,
    bins=None,
    alphas=None,
    gff_legend=False,
    prefix='',
):
    """
    create figure color legend
    """
    max_legend_length_row = 8
    max_legend_length_col = 4

    # Handle default value for bins
    if bins is None:
        bins = []

    # define output file
    if filetype not in ['png', 'pdf', 'svg']:
        text = 'Provide valid file type - png, pdf, or svg'
        logging.info(text)
        filetype = 'png'

    # check if length of information fit
    if not gff_legend and (
        (bins != [] and len(colors) != lcs_shading_num + 1)
        or (bins != [] and len(colors) != len(bins) + 1)
    ):
        if bins != [] and len(colors) != lcs_shading_num + 1:
            text = (
                '**Attention**\nlcs_shading_num (%d) does not match number of colors (%d)!\n'
                % (lcs_shading_num, len(bins))
            )
        elif bins != [] and len(colors) != len(bins) + 1:
            text = (
                '**Attention**\nnumber of LCS length bins (%d) does not match number of colors (%d)!\n'
                % (len(colors), len(bins))
            )
        logging.info(text)
    elif gff_legend and len(bins) != len(colors):
        text = (
            '**Attention**\nnumber of GFF Feature Types (%d) does not match number of colors (%d)!\n'
            % (len(colors), len(bins))
        )
        logging.info(text)

    # set alpha values to opaque if none are provided
    if alphas is None:
        alphas = []
        alphas = [1] * len(colors)

    # legend data points
    data_points = list(range(len(colors)))
    if not gff_legend:
        # specify intervals, if max_len provided
        if max_len is not None:
            _multi_factor = 100  # one digit
            if max_len <= 1:
                _multi_factor = 1000  # two digits
            # len_interval_size = (max_len-min_len) * multi_factor *1. // lcs_shading_num * (1./ multi_factor)
            len_interval_size = (max_len - min_len) * 1.0 / lcs_shading_num
            len_pos = [float('%.2f' % (min_len))]
            # calculate interval positions
            len_pos.extend(
                [
                    float('%.2f' % (len_pos[-1] + len_interval_size))
                    for _ in range(lcs_shading_num)
                ]
            )

            if prefix.startswith('custom-matrix') and (
                0 <= max_len <= 100 and 0 <= min_len <= 100
            ):
                unit = '%'
            elif prefix.startswith('custom-matrix'):
                unit = ''

            text = (
                '\n%d Legend intervals from %.2f to %.2f: \n\t%s - number: %d, step: %.2f, unit: %s\n'
                % (
                    lcs_shading_num + 1,
                    min_len,
                    max_len,
                    str(len_pos),
                    len(len_pos),
                    len_interval_size,
                    unit,
                )
            )
            logging.info(text)
            pos = len_pos
            interval_size = len_interval_size
        # generate legend labels acc. to standard interval notation
        else:
            # use default max_len = 100 and min_len = 0
            len_interval_size = 100.0 / lcs_shading_num
            pos = [float('%.2f' % (0))]
            # calculate interval positions
            pos.extend(
                [
                    float('%.2f' % (pos[-1] + len_interval_size))
                    for _ in range(lcs_shading_num)
                ]
            )

            # interval_size = 100 // lcs_shading_num
            # pos = range(interval_size, 101+interval_size, interval_size)

        # remove unneccessary zeros in decimal places (i.e. if x.x00 in all entries)
        while True:
            last_digit_all_zero = True
            no_delim = False
            for idx in range(len(pos)):
                # only process if fraction with decimal places
                if '.' not in str(pos[idx]):
                    no_delim = True
                    break
                # only process when all entries end in zero
                elif str(pos[idx])[-1] != '0':
                    last_digit_all_zero = False
                    break
            if not last_digit_all_zero or no_delim:
                break
            # remove last decimal place (== 0) from all entries
            else:
                temp_pos = pos[:]
                for idx in range(len(pos)):
                    if not str(pos[idx])[-2] == '.':
                        pos[idx] = float(str(pos[idx])[:-1])
                    else:
                        pos[idx] = int(str(pos[idx])[:-2])
                logging.info('Shortening legend entries: %s - %s' % (temp_pos, pos))

        # eliminate fractions if unit == bp/aa
        if unit in ['aa', 'bp']:
            for idx in range(len(pos)):
                temp_pos = pos[:]
                rounded_unit = False
                if '.' in str(pos[idx]):
                    rounded_unit = True
                    # round values up to next integer (keep integer, if not a fraction)
                    pos[idx] = int(pos[idx] / 1) + int(pos[idx] % 1 > 0)
                    if idx == len(pos) - 1 and pos[idx] == 101:
                        pos[idx] = 100
            if rounded_unit:
                logging.info(
                    "Fractions not permitted for unit '%s': %s -> %s"
                    % (unit, temp_pos, pos)
                )

        if bins != []:  # labels provided
            legend_labels = bins[:]
            legend_labels.append('max')
            legend_labels_lengths = []
            for item in bins:
                legend_labels_lengths.append(
                    '[%d %s, %d %s)' % (item - min(bins), unit, item, unit)
                )
            if len(bins) == len(colors) - 1:
                legend_labels_lengths.append(
                    '[%d %s, %s]' % (max(bins), unit, '\u221e')
                )  # infinite

        else:
            legend_labels = []
            legend_labels_lengths = []
            for idx in range(len(pos)):
                num = pos[idx]
                try:
                    legend_labels.append('[%d%%, %d%%)' % (num - interval_size, num))
                except Exception:
                    legend_labels.append(
                        '[%d%%, %d%%)' % (num - len_interval_size, num)
                    )
                if max_len is not None:
                    num = len_pos[idx]
                    # as int or float
                    if num == int(num) and int(len_interval_size) == len_interval_size:
                        legend_labels_lengths.append(
                            '[%d %s, %d %s)'
                            % (num, unit, num + len_interval_size, unit)
                        )
                    else:
                        legend_labels_lengths.append(
                            '[%.2f %s, %.2f %s)'
                            % (num, unit, num + len_interval_size, unit)
                        )
            legend_labels[-1] = '100' + unit
            if max_len is not None:
                if num == int(num) and int(len_interval_size) == len_interval_size:
                    legend_labels_lengths[-1] = '[%d %s, \u221e]' % (max_len, unit)
                else:
                    legend_labels_lengths[-1] = '[%.2f %s, \u221e]' % (max_len, unit)

    # set labels and choose file name
    if gff_legend:
        label_text = bins[:]
        edge_col = None
        legend_file_name = 'GFF_Shading_Legend_n%d.' % lcs_shading_num + filetype
    elif max_len is not None:
        label_text = legend_labels_lengths[:]
        edge_col = 'black'
        legend_file_name = (
            'Polydotplot_LCS_Shading_Legend_max%d%s_n%d.'
            % (max_len, unit, lcs_shading_num + 1)
            + filetype
        )
    elif bins != []:
        label_text = legend_labels_lengths[:]
        edge_col = 'black'
        legend_file_name = (
            'Polydotplot_LCS_Shading_Legend_%d%s_n%d.'
            % (bins[0], unit, lcs_shading_num + 1)
            + filetype
        )
    else:
        label_text = legend_labels[:]
        edge_col = 'black'
        legend_file_name = (
            'Polydotplot_LCS_Shading_Legend_%%len_n%d.' % (lcs_shading_num + 1)
            + filetype
        )

    if prefix is not None and prefix != '':
        if not prefix.endswith('-'):
            prefix = prefix + '-'
        legend_type = 'LCS'
        if prefix.startswith('custom-matrix'):
            prefix = prefix.replace('custom-matrix', '')[1:]
            legend_type = 'CustomMatrix'
        legend_file_name = prefix + legend_file_name.replace('LCS', legend_type)

    # plot legend figure
    fig, ax = P.subplots(3, 1, figsize=(len(colors) * 2, len(colors) * 2))
    for idx in range(len(colors)):
        ax[0].bar(
            data_points[idx] + 1,
            data_points[idx] + 1,
            color=colors[idx],
            label=label_text[idx],
            alpha=alphas[idx],
            edgecolor=edge_col,
        )
        ax[1].bar(
            data_points[idx] + 1,
            0,
            color=colors[idx],
            label=label_text[idx],
            alpha=alphas[idx],
            edgecolor=edge_col,
        )
        ax[2].bar(
            data_points[idx] + 1,
            0,
            color=colors[idx],
            label=label_text[idx],
            alpha=alphas[idx],
            edgecolor=edge_col,
        )
    ax[1].set_ylim(0, 1)
    ax[2].set_ylim(0, 1)
    ax[1].legend(
        ncol=((len(colors) - 1) // max_legend_length_row) + 1, framealpha=1
    )  # vertical legend
    col_num = len(colors)
    if len(colors) > max_legend_length_col:
        remainder = 0
        if len(colors) % max_legend_length_col != 0:
            remainder = 1
        row_num = len(colors) // max_legend_length_col + remainder
        remainder = 0
        if len(colors) % row_num != 0:
            remainder = 1
        col_num = len(colors) // row_num + remainder
    ax[2].legend(ncol=col_num, framealpha=1)  # horizontal legend

    P.savefig(legend_file_name)

    return legend_file_name



import logging
import os
import sys

from flexidot._version import __version__


def check_kmer_length(kmer_length):
    """
    Logs a warning if the kmer length is less than 10.

    Parameters:
    kmer_length (int): The kmer length to check.
    """
    if kmer_length < 10:
        logging.warning(
            'Kmer length is less than 10. This may result in less accurate dotplots.'
        )


def check_input(filename):
    """
    Check if input files exist
    """
    # Check if the input file exists
    if not os.path.isfile(filename):
        logging.error('Input file does not exist. Quitting.')
        raise FileNotFoundError(f"Input sequence file '{filename}' does not exist.")
    sys.stderr.write(f'\033[92m  Input file found: {filename}\033[0m\n')


def check_output_dir(outdir):
    """
    Check if the output directory exists
    """
    # Check if the output directory exists
    if not os.path.isdir(outdir):
        logging.error('Output directory does not exist. Quitting.')
        raise FileNotFoundError(f"Output directory '{outdir}' does not exist.")
    sys.stderr.write(f'\033[92m  Output directory found: {outdir}\033[0m\n\n')


def print_summary(args):
    """
    Print a summary of the selected options to stderr.
    Verifies the existence of any provided filenames.
    """
    # Define color for headings
    heading_color = '\033[1;34m'
    reset_color = '\033[0m'

    # Log command line arguments
    sys.stderr.write(f'\n{heading_color}CMD line input{reset_color}\n')
    sys.stderr.write('  ${0}\n\n'.format(' '.join(sys.argv)))

    # Input/output arguments
    sys.stderr.write(f'{heading_color}Input/Output Arguments:{reset_color}\n')
    sys.stderr.write(f'  Input files [-i]: {args.infiles}\n')
    for infile in args.infiles:
        check_input(infile)
    sys.stderr.write(f'  Output prefix [-o]: {args.output_prefix}\n')
    sys.stderr.write(f'  Output directory [--outdir]: {args.outdir}\n')
    check_output_dir(args.outdir)

    # Collage arguments
    sys.stderr.write(f'{heading_color}Collage Arguments:{reset_color}\n')
    sys.stderr.write(f'  Collage [-c]: {args.collage}\n')
    sys.stderr.write(f'  Number of columns per page [--n_col]: {args.n_col}\n')
    sys.stderr.write(f'  Number of rows per page [--n_row]: {args.n_row}\n\n')

    # File type
    sys.stderr.write(f'{heading_color}File Type:{reset_color}\n')
    sys.stderr.write(f'  Output file format [-f]: {args.filetype}\n\n')

    # Sorting
    sys.stderr.write(f'{heading_color}Sorting:{reset_color}\n')
    sys.stderr.write(f'  Sort sequences alphabetically [-s]: {args.sort}\n\n')

    # Calculation parameters
    sys.stderr.write(f'{heading_color}Calculation Parameters:{reset_color}\n')
    sys.stderr.write(f'  Wordsize (kmer length) [-k]: {args.wordsize}\n')
    mode_descriptions = {
        '0': 'self [default]',
        '1': 'paired',
        '2': 'poly (matrix with all-against-all dotplots)',
    }
    modes = [mode_descriptions[mode] for mode in args.mode]
    sys.stderr.write(f'  Modes [-m]: {", ".join(modes)}\n')
    sys.stderr.write(f'  Sequence type [-t]: {args.type_seq}\n')
    sys.stderr.write(f'  Wobble conversion [-w]: {args.wobble_conversion}\n')
    sys.stderr.write(f"  Maximum 'N' percentage [-n_max]: {args.max_n}%\n")
    sys.stderr.write(f'  Substitution count [-S]: {args.substitution_count}\n')
    sys.stderr.write(f'  No reverse complementary matches [-r]: {args.norev}\n')
    sys.stderr.write(f'  Only vs first sequence [-O]: {args.only_vs_first_seq}\n\n')

    # Graphic formatting
    sys.stderr.write(f'{heading_color}Graphic Formatting:{reset_color}\n')
    sys.stderr.write(f'  Line width [-A]: {args.line_width}\n')
    sys.stderr.write(f'  Line color (forward) [-B]: {args.line_col_for}\n')
    sys.stderr.write(f'  Line color (reverse) [-C]: {args.line_col_rev}\n')
    sys.stderr.write(f'  X-label position [-D]: {args.x_label_pos}\n')
    sys.stderr.write(f'  Font size [-E]: {args.label_size}\n')
    sys.stderr.write(f'  Spacing between dotplots [-F]: {args.spacing}\n')
    sys.stderr.write(f'  Length scaling [-L]: {args.length_scaling}\n')
    sys.stderr.write(f'  Mirror y-axis [-M]: {args.mirror_y_axis}\n')
    sys.stderr.write(f'  Plot size [-P]: {args.plot_size}\n')
    representation_descriptions = {0: 'full [default]', 1: 'upper', 2: 'lower'}
    sys.stderr.write(
        f'  Representation [-R]: {representation_descriptions[args.representation]}\n'
    )
    sys.stderr.write(f'  Title length [-T]: {args.title_length}\n\n')

    # GFF shading
    sys.stderr.write(f'{heading_color}GFF Shading:{reset_color}\n')
    sys.stderr.write(f'  GFF files [-g]: {args.gff}\n')
    if args.gff:
        for gff_file in args.gff:
            check_input(gff_file)
    if args.gff_color_config:
        sys.stderr.write(f'  GFF color config [-G]: {args.gff_color_config}\n')
        check_input(args.gff_color_config)
        sys.stderr.write('\n')

    # Longest Common Subsequence (LCS) shading
    sys.stderr.write(
        f'\n{heading_color}Longest Common Subsequence (LCS) Shading:{reset_color}\n'
    )
    sys.stderr.write(f'  LCS shading [-x]: {args.lcs_shading}\n')
    sys.stderr.write(f'  LCS shading intervals [-X]: {args.lcs_shading_num}\n')
    lcs_shading_ref_descriptions = {
        0: 'maximal LCS length [default]',
        1: 'maximally possible length',
        2: 'given interval sizes',
    }
    sys.stderr.write(
        f'  LCS shading reference [-y]: {lcs_shading_ref_descriptions[args.lcs_shading_ref]}\n'
    )
    sys.stderr.write(
        f'  LCS shading interval length [-Y]: {args.lcs_shading_interval_len}\n'
    )
    lcs_shading_ori_descriptions = {
        0: 'forward [default]',
        1: 'reverse',
        2: 'both strands',
    }
    sys.stderr.write(
        f'  LCS shading orientation [-z]: {lcs_shading_ori_descriptions[args.lcs_shading_ori]}\n\n'
    )

    # Custom matrix shading
    sys.stderr.write(f'{heading_color}Custom Matrix Shading:{reset_color}\n')
    sys.stderr.write(f'  User matrix file [-u]: {args.user_matrix_file}\n')
    if args.user_matrix_file:
        check_input(args.user_matrix_file)
    sys.stderr.write(f'  User matrix print [-U]: {args.user_matrix_print}\n\n')

    # Other options
    sys.stderr.write(f'{heading_color}Other Options:{reset_color}\n')
    sys.stderr.write(f'  Logging level [--loglevel]: {args.loglevel}\n')
    sys.stderr.write(f'  Log file [--logfile]: {args.logfile}\n')
    sys.stderr.write(f'  Version [-v]: {__version__}\n\n')



###############################
#     Analysis Functions      #
###############################

import logging

import numpy as np


def wobble_replacement(sequence, general_ambiguity_code):
    """
    get all degenerated sequences for sequence with ambiguous residues
    (only residues considered that are keys in wobble_dictionary)
    """

    # get positions of ambiguous residues
    wobble_pos = []
    for idx in range(len(sequence)):
        letter = sequence[idx]
        if letter in list(general_ambiguity_code.keys()):
            wobble_pos.append(idx)

    text = '\t%d wobbles' % len(wobble_pos)
    logging.debug(text)

    # replace one wobble through each iteration by all possible residues
    # repeat if still wobbles in new kmers
    kmer_variants = [sequence]
    while True:
        text = '\t\t%d kmer variants' % len(kmer_variants)
        logging.debug(text)
        temp_kmers = set()
        for kmer in kmer_variants:
            for idx in wobble_pos:
                letter = kmer[idx]
                if letter in list(general_ambiguity_code.keys()):
                    for base in general_ambiguity_code[kmer[idx]]:
                        newkmer = kmer[:idx] + base + kmer[idx + 1 :]
                        temp_kmers.add(newkmer)
        wobble = False
        for kmer in temp_kmers:
            for idx in range(len(kmer)):
                letter = kmer[idx]
                if letter in list(general_ambiguity_code.keys()):
                    wobble = True
                    break
            if wobble:
                break
        kmer_variants = set(list(temp_kmers)[:])
        if not wobble:
            break

    return kmer_variants


def split_diagonals(data, stepsize=1):
    """
    split array if point difference exceeds stepsize
    data = sorted list of numbers
    """
    return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)


def longest_common_substring(s1, s2):
    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    longest, _x_longest = 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    _x_longest = x
            else:
                m[x][y] = 0
    return longest


def lcs_from_x_values(x_values):
    """
    calculate length of longest common substring based on nested list of numbers
    """
    if len(x_values) == 0:
        return 0
    # get lengths of each subarray data
    lengths = np.array([len(i) for i in x_values])
    return max(lengths)



"""
Logging configuration for the Flexidot package.

This module provides functionality to initialize and configure logging with rich
formatting for better readability in terminal output. It uses the 'rich' library
to create visually enhanced log messages.
"""

import logging
from typing import Optional

from rich.console import Console
from rich.logging import RichHandler


def init_logging(loglevel: str = 'DEBUG', logfile: Optional[str] = None) -> None:
    """
    Initialize root logger with specified log level and rich formatting.

    Configures the global logging system with rich formatting for all modules
    that use the standard logging calls.

    Parameters
    ----------
    loglevel : str, optional
        The log level to use (e.g., "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        by default "DEBUG".
    logfile : str, optional
        Path to a file where logs should be written. If provided, logging output
        will be sent to both the console and this file. If None, logs will only
        be displayed in the console, by default None.

    Returns
    -------
    None
        This function configures the global logging system and doesn't return a value.

    Raises
    ------
    ValueError
        If the provided log level is invalid.
    """
    # Convert log level string to numeric value
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f'Invalid log level: {loglevel}')

    # Get the root logger
    root_logger = logging.getLogger()

    # Clear existing handlers if any are present
    if root_logger.hasHandlers():
        root_logger.handlers.clear()

    # Add rich handler that outputs to stderr for better visibility in scripts
    root_logger.addHandler(RichHandler(console=Console(stderr=True)))

    # Add file handler if logfile is specified
    if logfile:
        file_handler = logging.FileHandler(logfile)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)

    # Set the logger's level according to specified loglevel
    root_logger.setLevel(numeric_level)



import logging
import time

from colormap import rgb2hex
from colour import Color
from matplotlib import colormaps


# TODO: Remove internal logging message. Return delta time instead of printing it.
# Check if "now" value is used elsewhere.
def time_track(starting_time, show=True):
    """
    calculate time passed since last time measurement
    """
    now = time.time()
    delta = now - starting_time
    if show:
        logging.info(f'{delta} seconds')
    return delta


def calc_fig_ratio(ncols, nrows, plot_size):
    """
    Calculate size ratio for given number of columns (ncols) and rows (nrows)
    with plot_size as maximum width and length
    """
    ratio = ncols * 1.0 / nrows
    logging.debug(' '.join([str(ncols), str(nrows), str(ratio)]))
    if ncols >= nrows:
        figsize_x = plot_size
        figsize_y = plot_size / ratio
    else:
        figsize_x = plot_size * ratio
        figsize_y = plot_size
    return figsize_x, figsize_y


def shorten_name(seq_name, max_len=20, title_clip_pos='B'):  # , delim="_"):
    """
    shorten sequence names (for diagram titles)
    """

    if len(seq_name) <= max_len:
        return seq_name

    # take last characters
    if title_clip_pos == 'E':
        name = seq_name[len(seq_name) - max_len :]

    # take first characters
    else:
        name = seq_name[:max_len]

    """# keep first and last part if multiple parts separated by delimiter (e.g. species_prefix + sequence_id)
    if delim in seq_name:
        if seq_name.count(delim) >= 2:
            name = "%s..." % delim.join(seq_name.split(delim)[:1]) + seq_name.split(delim)[-1] # .replace("_000", "-")
        else:
            name = seq_name[:((max_len-2)//2)] + "..." + seq_name[((max_len-2)//2):]

        if len(name) > max_len:
            name = name[:((max_len-2)//2)] + "..." + name[((max_len-2)//2):]
    else:
        name = seq_name[:((max_len-2)//2)] + "..." + seq_name[((max_len-2)//2):]
    """

    return name


def unicode_name(name):
    """
    replace non-ascii characters in string (e.g. for use in matplotlib)
    """
    unicode_string = eval('u"%s"' % name)
    return ''.join(char for char in unicode_string if ord(char) < 128)
    # return unicodedata.normalize("NFKD", unicode_string).encode("ascii", "ignore")


def create_color_list(
    number: int, color_map: str = 'Greys', max_grey: str = '#595959'
) -> list:
    """
    Create color list with given number of entries
    grey by default, matplotlib color_map can be provided
    """
    if color_map not in list(colormaps):
        logging.warning(
            f'Invalid color_map {color_map} provided! - Examples: {list(colormaps)}.'
        )
        logging.warning('See https://matplotlib.org/users/colormaps.html\n')

    try:
        # create pylab colormap
        cmap = eval('P.cm.' + color_map)

        # get descrete color list from pylab
        cmaplist = [cmap(i) for i in range(cmap.N)]  # extract colors from map

        # determine positions for number of colors required
        steps = round((len(cmaplist) - 1) / (number))
        numbers = list(range(0, len(cmaplist), int(steps)))

        # extract RGB color and convert to hex code
        colors = []
        for idx in numbers[:-1]:
            rgba_color = cmaplist[idx]
            rgb_color = rgba_color[:3]
            col = rgb2hex(
                int(rgb_color[0] * 255),
                int(rgb_color[1] * 255),
                int(rgb_color[2] * 255),
            )
            colors.append(col)

    # Default to Greys color scheme if an error occurs
    except Exception as e:
        logging.warning(f'An error occurred: {e}')
        logging.warning('Using grey color scheme instead.')

        colors = list(Color('#FFFFFF').range_to(Color(max_grey), number))  # grey
        for idx in range(len(colors)):
            colors[idx] = str(colors[idx]).replace('Color ', '')
            if '#' in colors[idx] and len(colors[idx]) != 7:
                # print colors[idx]
                colors[idx] = colors[idx] + colors[idx][-(7 - len(colors[idx])) :]

    logging.info('%d Colors: %s' % (len(colors), ', '.join(colors)))

    if len(colors) < number:
        logging.info(
            '\nError in color range definition! %d colors missing\n'
            % (number - len(colors))
        )

    return colors



###############################
#     Matching Functions      #
###############################

from collections import defaultdict
import logging
from typing import Tuple, Union

from Bio.Seq import Seq
import numpy as np
import regex

# from flexidot.utils.utils import time_track
from flexidot.utils.alphabets import alphabets
from flexidot.utils.analysis import (
    lcs_from_x_values,
    split_diagonals,
    wobble_replacement,
)


def find_match_pos_diag(
    seq1: Union[str, Seq],
    seq2: Union[str, Seq],
    wordsize: int,
    report_lcs: bool = False,
    rc_option: bool = True,
    convert_wobbles: bool = False,
    max_N_percentage: float = 0,
    type_nuc: bool = True,
) -> Union[
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, int, int],
]:
    """
    Find all matching positions with matches >= wordsize
    convert matching points into lines of the length of the match
    (+ optional handling of ambiguities)
    """
    # TODO: check if seq1 and se2 are the same sequence (e.g. in case of self-alignment)
    # If so, then can skip counting the fwd kmers for seq2 and recycle the results from seq1
    if seq1 == seq2:
        logging.debug('Self-alignment detected. Recycling results from seq1 for seq2.')
        self_alignment = True
    else:
        self_alignment = False

    # Look for Ns in DNA or Xs in proeins (minimum word size)
    if type_nuc:
        unknown_residue = 'N'
    else:
        unknown_residue = 'X'

    # Calculate the maximum number of Ns allowed in a kmer
    max_N_count = (max_N_percentage / 100.0) * wordsize

    # read sequences
    seq_one = seq1.upper()
    # len_one = len(seq_one)
    seq_two = seq2.upper()
    len_two = len(seq_two)

    # Set ambiguity code for wobble replacement
    general_ambiguity_code = alphabets(type_nuc)[
        2
    ]  # nucleotide_ambiguity_code or aminoacid_ambiguity_code

    # Forward
    #################################
    kmer_pos_dict_one = defaultdict(list)  # Seq1
    kmer_pos_dict_two = defaultdict(list)  # Seq2

    # Reverse complement
    #################################
    kmer_pos_dict_three = defaultdict(list)  # Seq1
    kmer_pos_dict_four = defaultdict(list)  # Seq2

    # Create dictionaries to index kmer (wordsize) positions in the sequence
    if self_alignment and rc_option:
        # Compare seq1 forward to self and self reverse
        data_list = [
            (str(seq_one), kmer_pos_dict_one),
            (str(seq_one.reverse_complement()), kmer_pos_dict_three),
        ]
    elif self_alignment and not rc_option:
        # Compare seq1 forward to self only
        data_list = [
            (str(seq_one), kmer_pos_dict_one),
        ]
    elif rc_option:
        # Compare seq1 forward to seq2 forward and reverse
        data_list = [
            (str(seq_one), kmer_pos_dict_one),
            (str(seq_two), kmer_pos_dict_two),
            # (str(seq_one), kmer_pos_dict_three), #TODO: Check if this is needed
            (str(seq_two.reverse_complement()), kmer_pos_dict_four),
        ]
    else:
        # Compare seq1 and seq2 forward only
        data_list = [
            (str(seq_one), kmer_pos_dict_one),
            (str(seq_two), kmer_pos_dict_two),
        ]

    # Step through each sequence and add kmers to dictionary
    for seq, kmer_pos_dict in data_list:
        # Track number of kmers skipped due to Ns > max_N_count
        skipped_Ns = 0
        # Step through sequence and add kmer positions to dictionary
        for i in range(len(seq) - wordsize + 1):
            # Extract kmer
            kmer = seq[i : i + wordsize]
            # Count Ns in kmer
            Ns_in_kmer = kmer.count(unknown_residue)
            # Discard kmer, if too many Ns included
            if Ns_in_kmer <= max_N_count:
                if not convert_wobbles:
                    if Ns_in_kmer == 0:
                        # Add kmer to dictionary if no Ns.
                        kmer_pos_dict[kmer].append(i)
                    else:
                        # Skip kmers with Ns
                        skipped_Ns += 1
                else:  # Deal with ambiguous characters
                    # Set as True if any wobble characters are present in the kmer
                    has_wobbles = any(
                        item in kmer for item in general_ambiguity_code.keys()
                    )
                    if not has_wobbles:
                        kmer_pos_dict[kmer].append(i)
                    else:
                        # Replace wobble characters with all possible variants
                        kmer_variants = wobble_replacement(kmer, general_ambiguity_code)
                        for new_kmer in kmer_variants:
                            kmer_pos_dict[new_kmer].append(i)
            else:
                skipped_Ns += 1

        # Log number of skipped kmers
        if skipped_Ns > 0:
            if convert_wobbles:
                logging.debug(
                    'Skipped %i kmers due to {unknown_residue}s > %i'
                    % (skipped_Ns, max_N_count)
                )
            else:
                logging.debug(
                    'Skipped %i kmers containing {unknown_residue}s' % (skipped_Ns)
                )

    # If self alignment, duplicate self fwd and rev dictionaries
    if self_alignment:
        # If self alignment, copy kmer_pos_dict_one to kmer_pos_dict_two
        kmer_pos_dict_two = kmer_pos_dict_one.copy()
        # Copy kmer_pos_dict_three to kmer_pos_dict_four
        kmer_pos_dict_four = kmer_pos_dict_three.copy()

    # Find kmers shared between both sequences in forward orientation
    matches_for = set(kmer_pos_dict_one).intersection(kmer_pos_dict_two)

    # Find kmers shared between Seq1 forward and Seq2 reverse orientation
    matches_rc = set(kmer_pos_dict_one).intersection(kmer_pos_dict_four)

    # TODO: Check if above is correct. Do we gain any extra kmers over set(matches_for, matches_rc) if we also check the seq1_rc?
    # print("matches_for: ", type(matches_for) ,matches_for)
    # print("matches_rc: ", matches_rc)

    logging.debug(
        '[matches: %i forward; %i reverse]' % (len(matches_for), len(matches_rc))
    )

    # Create lists of x and y coordinates for scatter plot
    # Keep all coordinates of all shared kmers (may match multiple times)
    diag_dict_for = defaultdict(set)
    diag_dict_rc = defaultdict(set)

    for match_list, pos_dict1, pos_dict2, diag_dict in [
        (matches_for, kmer_pos_dict_one, kmer_pos_dict_two, diag_dict_for),
        (matches_rc, kmer_pos_dict_one, kmer_pos_dict_four, diag_dict_rc),
    ]:
        for kmer in match_list:
            for i in pos_dict1[kmer]:
                for j in pos_dict2[kmer]:
                    diag = i - j
                    points = set(range(i + 1, i + wordsize + 1))
                    diag_dict[diag].update(points)

    # Convert coordinate points to line start and stop positions
    x1 = []  # x values reverse
    y1 = []  # y values forward
    for diag in list(diag_dict_for.keys()):
        x_values = np.array(sorted(diag_dict_for[diag]))
        x1.extend(split_diagonals(x_values))
        y_values = split_diagonals(x_values - diag)
        y1.extend(y_values)

    x2 = []  # x values rc
    y2 = []  # y values rc
    if rc_option:
        for diag in list(diag_dict_rc.keys()):
            factor = len_two + diag + 1
            x_values = np.array(sorted(diag_dict_rc[diag]))
            x2.extend(split_diagonals(x_values))
            y_values = split_diagonals(factor - x_values, -1)
            y2.extend(y_values)

    if not report_lcs:
        return (
            # x values forward matches
            np.array([np.array(x) for x in x1], dtype=object),
            # y values forward matches
            np.array([np.array(y) for y in y1], dtype=object),
            # x values rc (ascending)
            np.array([np.array(x) for x in x2], dtype=object),
            # y values rc (descending)
            np.array([np.array(y) for y in y2], dtype=object),
        )
    else:
        # Get length of longest common substring based on match lengths
        lcs_for = lcs_from_x_values(x1)
        lcs_rev = lcs_from_x_values(x2)
        return (
            np.array([np.array(x) for x in x1], dtype=object),
            np.array([np.array(y) for y in y1], dtype=object),
            np.array([np.array(x) for x in x2], dtype=object),
            np.array([np.array(y) for y in y2], dtype=object),
            lcs_for,
            lcs_rev,
        )


def find_match_pos_regex(
    seq1,
    seq2,
    wordsize,
    substitution_count=0,
    report_lcs=False,
    rc_option=True,
    convert_wobbles=False,
    max_N_percentage=0,
    type_nuc=True,
):
    """
    find all matching positions with matches >= wordsize via regular expression search
    fuzzy matching - allow up to substitution_count substitutions
    convert matching points into lines of the length of the match
    (+ optional handling of ambiguities)
    """

    # read sequences
    seq_one = seq1.upper()
    # len_one = len(seq_one)
    seq_two = seq2.upper()
    len_two = len(seq_two)

    # set ambiguity code for wobble replacement
    general_ambiguity_code = alphabets(type_nuc)[
        2
    ]  # nucleotide_ambiguity_code or aminoacid_ambiguity_code
    ambiguity_match_dict = alphabets(type_nuc)[3]

    ambiq_residues = '[%s]' % ''.join(list(general_ambiguity_code.keys()))

    # look for Ns in DNA or Xs in proeins (minimum word size)
    if type_nuc:
        any_residue = 'N'
    else:
        any_residue = 'X'

    # check for wobble presence
    if not (
        regex.search(ambiq_residues, str(seq_one)) is None
        and regex.search(ambiq_residues, str(seq_two)) is None
    ):
        wobble_found = True
    else:
        wobble_found = False

    # dictionary for matches
    diag_dict_for = {}
    diag_dict_rc = {}
    counter = [0, 0]

    # one-way matching
    if rc_option:
        data_list = [
            (str(seq_one), str(seq_two), diag_dict_for, 0),
            (str(seq_one), str(seq_two.reverse_complement()), diag_dict_rc, 1),
        ]
    else:
        data_list = [(str(seq_one), str(seq_two), diag_dict_for, 0)]

    for seq_query, seq_target, diag_dict, counter_pos in data_list:
        # split query sequence into kmers
        if not rc_option and counter_pos == 1:
            break

        for idx in range(len(str(seq_query)) - wordsize + 1):
            kmer = str(seq_query)[idx : idx + wordsize]

            # skip excessive N/X stretches (big black areas)
            if kmer.count(any_residue) * 100.0 / wordsize <= max_N_percentage:
                #  convert kmer to regular expression for wobble_matching
                if convert_wobbles and wobble_found:
                    kmer_string = ''
                    # replace each residue with matching residues or wobbles
                    for jdx in range(len(kmer)):
                        kmer_string += ambiguity_match_dict[kmer[jdx]]
                else:
                    kmer_string = kmer

                # convert to regular expression tolerating substitution errors
                if type(substitution_count) is int and substitution_count != 0:
                    kmer_string = '(%s){s<=%d}' % (kmer_string, substitution_count)

                # search for regular expression in target sequence
                kdx = 0
                # start = True
                if regex.search(kmer_string, seq_target[kdx:]) is not None:
                    counter[counter_pos] += 1
                    while regex.search(kmer_string, seq_target[kdx:]) is not None:
                        # search for regular expression pattern in target sequence
                        result = regex.search(kmer_string, seq_target[kdx:])

                        kmer2 = seq_target[kdx:][result.start() : result.end()]

                        # skip excessive N/X stretches (big black areas)
                        if (
                            kmer2.count(any_residue) * 100.0 / wordsize
                            <= max_N_percentage
                        ):
                            diag = idx - (kdx + result.start())
                            points = set(range(idx + 1, idx + wordsize + 1))
                            if diag not in list(diag_dict.keys()):
                                diag_dict[diag] = points
                            else:
                                diag_dict[diag].update(points)

                        kdx += result.start() + 1
                        if kdx >= len(seq_target):
                            break
                        elif regex.search(kmer_string, seq_target[kdx:]) is not None:
                            counter[counter_pos] += 1

    text = '%5.i \tforward matches' % counter[0]
    text += '\n%5.i \treverse complementary matches' % counter[1]
    logging.debug(text)

    # convert coordinate points to line start and stop positions
    x1 = []  # x values reverse
    y1 = []  # y values forward
    for diag in list(diag_dict_for.keys()):
        x_values = np.array(sorted(diag_dict_for[diag]))
        x1.extend(split_diagonals(x_values))
        y_values = split_diagonals(x_values - diag)
        y1.extend(y_values)

    x2 = []  # x values rc
    y2 = []  # y values rc
    if rc_option:
        for diag in list(diag_dict_rc.keys()):
            factor = len_two + diag + 1
            x_values = np.array(sorted(diag_dict_rc[diag]))
            x2.extend(split_diagonals(x_values))
            y_values = split_diagonals(factor - x_values, -1)
            y2.extend(y_values)

    if not report_lcs:
        return (
            np.array([np.array(x) for x in x1], dtype=object),
            np.array([np.array(y) for y in y1], dtype=object),
            np.array([np.array(x) for x in x2], dtype=object),
            np.array([np.array(y) for y in y2], dtype=object),
        )
    else:
        # get length of longest common substring based on match lengths
        lcs_for = lcs_from_x_values(x1)
        lcs_rev = lcs_from_x_values(x2)
        return (
            np.array([np.array(x) for x in x1], dtype=object),
            np.array([np.array(y) for y in y1], dtype=object),
            np.array([np.array(x) for x in x2], dtype=object),
            np.array([np.array(y) for y in y2], dtype=object),
            lcs_for,
            lcs_rev,
        )



###############################
#     Dot Plot Functions      #
###############################

import logging

import matplotlib.collections as cllct
import matplotlib.gridspec as gridspec
import matplotlib.patches as patches
import numpy as np
import pylab as P

from flexidot.utils.file_handling import legend_figure, read_gffs, read_matrix, read_seq
from flexidot.utils.matching import find_match_pos_diag, find_match_pos_regex
from flexidot.utils.utils import (
    calc_fig_ratio,
    create_color_list,
    shorten_name,
    unicode_name,
)


def selfdotplot(
    input_fasta,
    wordsize,
    alphabetic_sorting=False,
    convert_wobbles=False,
    filetype='png',
    gff_color_dict=None,
    gff_files=None,
    label_size=10,
    line_col_for='#000000',  # defalut black
    line_col_rev='#009243',  # default green
    line_width=1,
    max_N_percentage=10,
    mirror_y_axis=False,
    multi=True,
    ncols=4,
    nrows=5,
    plot_size=10,
    prefix=None,
    substitution_count=0,
    title_clip_pos='B',
    title_length=float('Inf'),
    type_nuc=True,
):
    """
    self-against-self dotplot
    partially from biopython cookbook
    """

    # Initialize default gff_color_dict if None
    if gff_color_dict is None:
        gff_color_dict = {'others': ('grey', 1, 0)}

    # read sequences
    if gff_files is None:
        gff_files = []
    seq_dict, sequences = read_seq(input_fasta)
    if seq_dict == {}:
        logging.warning('Failed to load sequences.')
        return []

    if type_nuc:
        aa_bp_unit = 'bp'
    else:
        aa_bp_unit = 'aa'

    if alphabetic_sorting:
        sequences = sorted(sequences)

    # check if at least one input sequence
    if len(sequences) == 0:
        text = '\n%s\n\nCreating %s selfdotplot images\n%s\n\n=>' % (
            50 * '=',
            len(sequences),
            28 * '-',
        )
        text += ' No sequences provided for selfdotplot!\n\nTerminating polydotplot!'
        logging.info(text)
        return
    elif len(sequences) == 1 and multi:
        text = '\n\nCreating collage output for single selfdotplot!'
        text += "\nRecommendation: Change to individual mode by using '--collage_output n'!\n\n"
        logging.info(text)

    if multi and (ncols == 0 or nrows == 0):
        ncols = max(ncols, 1)
        nrows = max(nrows, 1)
        text = (
            '\n\nSelfdotplot Collage: Invalid collage - correcting number of rows and columns:\n\tncols=%d, nrows=%d\n'
            % (ncols, nrows)
        )
        logging.info(text)

    if multi and ncols > len(sequences):
        ncols = len(sequences)
        nrows = 1
        text = (
            '\n\nSelfdotplot Collage: Few sequences - correcting number of rows and columns:\n\tncols=%d, nrows=%d\n'
            % (ncols, nrows)
        )
        logging.info(text)
    elif multi and ncols * (nrows - 1) > len(sequences):
        nrows = ((len(sequences) - 1) // ncols) + 1
        text = (
            '\n\nSelfdotplot Collage: Few sequences - correcting number of rows:\n\tncols=%d, nrows=%d\n'
            % (ncols, nrows)
        )
        logging.info(text)

    if multi and not (nrows == 1 and ncols == 1) and plot_size <= label_size / 2:
        label_size = plot_size * 3 // 2
        text = 'Reducing label size for better visualization to %d\n' % label_size
        logging.info(text)

    # Initialize gff_files if None
    if gff_files is None:
        gff_files = []

    # read gff annotation data if provided for shading
    if gff_files:
        text = '\n%s\n\nReading %s GFF annotation files\n%s\n\n=> %s\n' % (
            50 * '=',
            len(gff_files),
            28 * '-',
            ', '.join(gff_files),
        )
        logging.info(text)
        if prefix:
            legend_prefix = prefix + '-Selfdotplot'
        else:
            legend_prefix = 'Selfdotplot'
        feat_dict = read_gffs(
            gff_files,
            color_dict=gff_color_dict,
            type_nuc=type_nuc,
            prefix=legend_prefix,
            filetype=filetype,
        )

    log_txt = '\n%s\n\nCreating %s selfdotplot images\n%s\n\n=>' % (
        50 * '=',
        len(sequences),
        28 * '-',
    )

    # preparations for file name
    name_graph = 'Selfdotplots'
    if prefix:
        if not prefix[-1] == '-':
            prefix = prefix + '-'
    else:
        prefix = ''
    suffix = ''
    if convert_wobbles:
        suffix += '_wobbles'
    if substitution_count != 0:
        suffix += '_S%d' % substitution_count
    if multi:
        suffix += '_collage'

    # calculate fig ratios
    if not multi:
        ncols = 1
        nrows = 1
    figsize_x, figsize_y = calc_fig_ratio(ncols, nrows, plot_size)

    P.cla()  # clear any prior graph
    if multi:
        P.figure(figsize=(figsize_x, figsize_y))
        page_counter = 1
    list_of_png_names = []

    counter = 0
    for seq_name in sequences:
        log_txt += '\n- ' + seq_name

        counter += 1
        if not multi:
            P.cla()  # clear any prior graph

        # read sequence
        seq_record = seq_dict[seq_name]
        name_seq = seq_record.id
        seq_one = seq_record.seq.upper()
        length_seq = len(seq_one)

        # get positions of matches
        if substitution_count != 0:
            # print "RE"
            x_lists, y_lists, x_lists_rc, y_lists_rc = find_match_pos_regex(
                seq_one,
                seq_one,
                wordsize,
                substitution_count=substitution_count,
                convert_wobbles=convert_wobbles,
                max_N_percentage=max_N_percentage,
                type_nuc=type_nuc,
            )
        else:
            # print "DIAG",
            x_lists, y_lists, x_lists_rc, y_lists_rc = find_match_pos_diag(
                seq_one,
                seq_one,
                wordsize,
                convert_wobbles=convert_wobbles,
                max_N_percentage=max_N_percentage,
                type_nuc=type_nuc,
            )

        # plotting with matplotlib
        #################################

        # combined plotting
        if multi:
            # plotting subplot with matplotlib
            ax = P.subplot(nrows, ncols, counter)  # rows, columns, plotnumber

            # shade annotated regions
            if gff_files:
                if seq_name in list(feat_dict.keys()):
                    features = feat_dict[seq_name]
                    for item in features:
                        feat_type, start, stop = item
                        feat_color, strength, zoom = gff_color_dict[feat_type.lower()]
                        start = max(0, start - zoom - 0.5)
                        stop = min(length_seq + 1, stop + zoom + 0.5)
                        width = stop - start
                        ax.add_patch(
                            patches.Rectangle(
                                (start, start),  # (x,y)
                                width,
                                width,  # width, height
                                edgecolor=None,
                                linewidth=line_width + zoom,
                                fill=True,
                                facecolor=feat_color,
                                alpha=strength,
                            )
                        )

            # collect lines
            lines = []
            color_list = []
            for x_lines, y_lines, col in [
                (x_lists_rc, y_lists_rc, line_col_rev),
                (x_lists, y_lists, line_col_for),
            ]:
                # If color is not white, add lines to plot
                if col != 'white':
                    for ldx in range(len(x_lines)):
                        lines.append(
                            [
                                (x_lines[ldx][0], y_lines[ldx][0]),
                                (x_lines[ldx][-1], y_lines[ldx][-1]),
                            ]
                        )
                        color_list.append(col)
            color_list = np.array(color_list)

            # draw lines
            lc = cllct.LineCollection(lines, colors=color_list, linewidths=line_width)
            ax.add_collection(lc)

            # format axes
            # print P.xticks()[0], P.yticks()[0]
            P.axis('scaled')  # make images quadratic
            P.xlim(0, length_seq + 1)
            if mirror_y_axis:
                P.ylim(0, length_seq + 1)  # rotate y axis (point upwards)
            else:
                P.ylim(length_seq + 1, 0)  # rotate y axis (point downwards)
            P.xlabel('[%s]' % aa_bp_unit, fontsize=label_size)
            P.ylabel('[%s]' % aa_bp_unit, fontsize=label_size)
            P.tick_params(axis='both', which='major', labelsize=label_size * 0.9)

            # # use same tick labels for x and y axis
            # tick_locs, tick_labels = P.yticks()
            # P.xticks(tick_locs)
            # P.xlim(0, length_seq+1)

            P.title(
                unicode_name(
                    shorten_name(
                        name_seq, max_len=title_length, title_clip_pos=title_clip_pos
                    )
                ),
                fontsize=label_size,
                fontweight='bold',
            )
            # P.title(unicode_name(name_seq), fontsize=label_size*1.3, fontweight='bold')

            # save figure and reinitiate if page is full
            if counter == ncols * nrows:
                # finalize layout - margins & spacing between plots
                try:
                    P.tight_layout(h_pad=0.02, w_pad=0.02)
                except Exception as e:
                    logging.info(
                        'Attention - pylab.tight_layout failed! Please check sequence names and layout settings! Error: %s'
                        % str(e)
                    )
                P.subplots_adjust(
                    hspace=0.5, wspace=0.5
                )  # space between rows - def 0.4

                # name and create output files (names derived from SEQNAME)
                fig_name = '%s%s_wordsize%i%s-%.3d.%s' % (
                    prefix,
                    name_graph,
                    wordsize,
                    suffix,
                    page_counter,
                    filetype,
                )
                P.savefig(fig_name, bbox_inches='tight')
                P.close()
                P.cla()

                list_of_png_names.append(fig_name)

                counter = 0
                page_counter += 1

                P.figure(figsize=(figsize_x, figsize_y))

        # plotting separate figure files
        else:  # not multi
            P.figure(figsize=(plot_size, plot_size))  # figure size needs to be a square
            ax = P.subplot(1, 1, 1)  # rows, columns, plotnumber

            # shade annotated regions
            if gff_files:
                if seq_name in list(feat_dict.keys()):
                    features = feat_dict[seq_name]
                    for item in features:
                        feat_type, start, stop = item
                        feat_color, strength, zoom = gff_color_dict[feat_type.lower()]
                        start = max(0, start - zoom - 0.5)
                        stop = min(length_seq + 1, stop + zoom + 0.5)
                        width = stop - start
                        ax.add_patch(
                            patches.Rectangle(
                                (start, start),  # (x,y)
                                width,
                                width,  # width, height
                                edgecolor=None,
                                linewidth=line_width + zoom,
                                fill=True,
                                facecolor=feat_color,
                                alpha=strength,
                            )
                        )

            # collect lines
            lines = []
            color_list = []
            for x_lines, y_lines, col in [
                (x_lists_rc, y_lists_rc, line_col_rev),
                (x_lists, y_lists, line_col_for),
            ]:
                # If color is not white, add lines to plot
                if col != 'white':
                    for ldx in range(len(x_lines)):
                        lines.append(
                            [
                                (x_lines[ldx][0], y_lines[ldx][0]),
                                (x_lines[ldx][-1], y_lines[ldx][-1]),
                            ]
                        )
                        color_list.append(col)

            color_list = np.array(color_list)

            # draw lines
            lc = cllct.LineCollection(lines, colors=color_list, linewidths=line_width)
            ax.add_collection(lc)

            # format axes
            P.axis('scaled')  # make images quadratic
            P.xlim(0, length_seq + 1)
            if mirror_y_axis:
                P.ylim(0, length_seq + 1)  # rotate y axis (point upwards)
            else:
                P.ylim(length_seq + 1, 0)  # rotate y axis (point downwards)
            P.xlabel('[%s]' % aa_bp_unit, fontsize=label_size)
            P.ylabel('[%s]' % aa_bp_unit, fontsize=label_size)
            P.tick_params(axis='both', which='major', labelsize=label_size * 0.9)

            # # use same tick labels for x and y axis
            # tick_locs, tick_labels = P.yticks()
            # P.xticks(tick_locs)
            # P.xlim(0, length_seq+1)

            P.title(
                unicode_name(
                    shorten_name(
                        name_seq, max_len=title_length, title_clip_pos=title_clip_pos
                    )
                ),
                fontsize=label_size * 1.3,
                fontweight='bold',
            )

            # name and create output files (names derived from SEQNAME)
            fig_name = '%s%s-%d_%s_wordsize%i%s.%s' % (
                prefix,
                name_graph,
                counter,
                shorten_name(
                    name_seq, max_len=title_length, title_clip_pos=title_clip_pos
                ),
                wordsize,
                suffix,
                filetype,
            )
            P.savefig(fig_name, bbox_inches='tight')

            P.close()
            P.cla()  # clear any prior graph

            list_of_png_names.append(fig_name)

    if multi and counter >= 1:
        # finalize layout - margins & spacing between plots
        try:
            P.tight_layout(h_pad=0.02, w_pad=0.02)
        except Exception:
            logging.info(
                'Attention - pylab.tight_layout failed! Please check sequence names and layout settings!'
            )
        P.subplots_adjust(hspace=0.5, wspace=0.5)  # space between rows - def 0.4

        # name and create output files (names derived from SEQNAME)
        fig_name = '%s%s_wordsize%i%s-%.3d.%s' % (
            prefix,
            name_graph,
            wordsize,
            suffix,
            page_counter,
            filetype,
        )
        P.savefig(fig_name, bbox_inches='tight')
        P.close()
        P.cla()  # clear any prior graph

        list_of_png_names.append(fig_name)

    log_txt += '\n\nDrawing selfdotplots done.\n'
    logging.info(log_txt)

    return list_of_png_names


def pairdotplot(
    input_fasta,
    wordsize,
    alphabetic_sorting=False,
    convert_wobbles=False,
    filetype='png',
    label_size=10,
    length_scaling=True,
    line_col_for='#000000',  # defalut black
    line_col_rev='#009243',  # default green
    line_width=1,
    max_N_percentage=10,
    mirror_y_axis=False,
    multi=True,
    ncols=4,
    nrows=5,
    only_vs_first_seq=False,
    plot_size=10,
    prefix=None,
    scale_delim_col='red',
    substitution_count=0,
    title_clip_pos='B',
    title_length=float('Inf'),
    type_nuc=True,
    x_label_pos_top=True,
):
    """
    pairwise dotplot (all-against-all)
    """

    # read sequences
    seq_dict, sequences = read_seq(input_fasta)
    if seq_dict == {}:
        logging.warning('Failed to load sequences.')
        return []

    if type_nuc:
        aa_bp_unit = 'bp'
    else:
        aa_bp_unit = 'aa'

    if alphabetic_sorting:
        sequences = sorted(sequences)

    # check if at least two input sequences
    if len(sequences) < 2:
        text = '\n%s\n\nCreating %d paired dotplot image \n%s\n\n=>' % (
            50 * '=',
            len(sequences) * (len(sequences) - 1) / 2,
            36 * '-',
        )
        text += ' Please provide at least two sequences for pairdotplot!\n\nTerminating paired dotplot!'
        logging.info(text)
        return
    elif len(sequences) == 2 and multi:
        text = '\n\nCreating collage output for single pairdotplot!'
        text += "\nRecommendation: Change to individual mode by using '--collage_output n'!\n\n"
        logging.info(text)

    if multi and (ncols == 0 or nrows == 0):
        ncols = max(ncols, 1)
        nrows = max(nrows, 1)
        text = (
            '\n\nPairdotplot Collage: Invalid collage settings - correcting number of rows and columns:\n\tncols=%d, nrows=%d\n'
            % (ncols, nrows)
        )
        logging.info(text)

    if multi and ncols > len(sequences) * (len(sequences) - 1):
        ncols = len(sequences)
        nrows = 1
        text = (
            '\n\nPairdotplot Collage: Few sequences - correcting number of rows and columns:\n\tncols=%d, nrows=%d\n'
            % (ncols, nrows)
        )
        logging.info(text)
    elif multi and ncols * (nrows - 1) > len(sequences) * (len(sequences) - 1):
        nrows = ((len(sequences) - 1) // ncols) + 1
        text = (
            '\n\nPairdotplot Collage: Few sequences - correcting number of rows:\n\tncols=%d, nrows=%d\n'
            % (ncols, nrows)
        )
        logging.info(text)

    if not only_vs_first_seq:
        text = '\n%s\n\nCreating %d paired dotplot image for\n%s\n\n=>' % (
            50 * '=',
            len(sequences) * (len(sequences) - 1) / 2,
            36 * '-',
        )
        text += ',\n'.join(sequences) + '\n'
    else:
        text = (
            "\n%s\n\nCreating %d paired dotplot images against 1st sequence '%s':\n%s\n\n=>"
            % (50 * '=', len(sequences) - 1, sequences[0], 36 * '-')
        )
        text += ',\n'.join(sequences[1:]) + '\n'
    logging.info(text)

    if multi and not (nrows == 1 and ncols == 1) and plot_size <= label_size / 2:
        label_size = plot_size * 3 // 2
        text = 'Reducing label size for better visualization to %d\n' % label_size
        logging.info(text)

    y_label_rotation = 'vertical'
    # for cartesian coordinate system with mirrored y-axis: plot x labels below plot
    if mirror_y_axis:
        x_label_pos_top = False

    # preparations for file name
    name_graph = 'Pairdotplot'
    if prefix:
        if not prefix[-1] == '-':
            prefix = prefix + '-'
    else:
        prefix = ''
    suffix = ''
    if convert_wobbles:
        suffix += '_wobbles'
    if substitution_count != 0:
        suffix += '_S%d' % substitution_count
    if length_scaling:
        suffix += '_scaled'
    if multi:
        suffix += '_collage'

    # calculate fig ratios
    if not multi:
        ncols = 1
        nrows = 1
    figsize_x, figsize_y = calc_fig_ratio(ncols, nrows, plot_size)

    P.cla()  # clear any prior graph
    list_of_png_names = []
    if multi:
        P.figure(figsize=(figsize_x, figsize_y))
        page_counter = 1

    # prepare LCS data file
    lcs_data_file = open(
        '%sPairdotplot_wordsize%d_lcs_data_file%s.txt'
        % (prefix, wordsize, suffix.replace('_scaled', '').replace('_collage', '')),
        'w',
    )
    lcs_data_file.write(
        '\t'.join(
            [
                '#title1',
                'title2',
                'len_seq1',
                'len_seq2',
                'len_lcs_for',
                '%_min_seq_len',
                'len_lcs_rev',
                '%_min_seq_len',
            ]
        )
        + '\n'
    )

    counter, seq_counter = 0, 0
    log_txt = '\nDrawing pairwise dotplots'

    seq_text = ''
    for idx in range(len(sequences) - 1):
        logging.debug('\n%d\t%s vs.' % ((seq_counter + 1), sequences[idx]))
        seq_text += '\n%d\t%s vs.' % ((seq_counter + 1), sequences[idx])

        rec_two = seq_dict[sequences[idx]]
        name_two = rec_two.id
        seq_two = rec_two.seq
        len_two = len(seq_two)

        for jdx in range(idx + 1, len(sequences)):
            rec_one = seq_dict[sequences[jdx]]
            name_one = rec_one.id
            seq_one = rec_one.seq
            len_one = len(seq_one)

            counter += 1
            seq_counter += 1

            logging.debug(sequences[jdx])
            seq_text += ' ' + sequences[jdx]

            if not seq_counter % 25:
                log_txt += ' ' + str(seq_counter)

            # get positions of matches
            if substitution_count != 0:
                # print "RE"
                x1, y1, x2, y2, lcs_for, lcs_rev = find_match_pos_regex(
                    seq_one,
                    seq_two,
                    wordsize,
                    substitution_count=substitution_count,
                    convert_wobbles=convert_wobbles,
                    max_N_percentage=max_N_percentage,
                    report_lcs=True,
                    type_nuc=type_nuc,
                )
            else:
                # print "DIAG"
                x1, y1, x2, y2, lcs_for, lcs_rev = find_match_pos_diag(
                    seq_one,
                    seq_two,
                    wordsize,
                    convert_wobbles=convert_wobbles,
                    max_N_percentage=max_N_percentage,
                    report_lcs=True,
                    type_nuc=type_nuc,
                )

            # write LCS data file
            lcs_data_file.write(
                '\t'.join(
                    [
                        name_one,
                        name_two,
                        str(len_one),
                        str(len_two),
                        str(lcs_for),
                        str(round((lcs_for * 100.0 / min(len_one, len_two)), 3)),
                        str(lcs_rev),
                        str(round((lcs_rev * 100.0 / min(len_one, len_two)), 3)),
                    ]
                )
                + '\n'
            )

            # Plotting with matplotlib
            #################################

            # combined plotting
            if multi:
                # plotting subplot with matplotlib
                ax = P.subplot(nrows, ncols, counter)  # rows, columns, plotnumber

            else:
                # calculate figure size for separate figures
                if len_one >= len_two:
                    pass
                    # sizing = (plot_size, max(2, (plot_size) * len_two * 1.0 / len_one))
                else:
                    pass
                    # sizing = (max(2, (plot_size) * len_one * 1.0 / len_two), plot_size)
                P.figure(figsize=(plot_size, plot_size))

                ax = P.subplot(1, 1, 1)

            # collect lines
            lines = []
            color_list = []
            for x_lines, y_lines, col in [
                (x2, y2, line_col_rev),
                (x1, y1, line_col_for),
            ]:
                # If color is not white, add lines to plot
                if col != 'white':
                    for ldx in range(len(x_lines)):
                        lines.append(
                            [
                                (x_lines[ldx][0], y_lines[ldx][0]),
                                (x_lines[ldx][-1], y_lines[ldx][-1]),
                            ]
                        )
                        color_list.append(col)
            color_list = np.array(color_list)

            # draw lines
            lc = cllct.LineCollection(lines, colors=color_list, linewidths=line_width)
            ax.add_collection(lc)

            # format axes
            P.xlabel(
                unicode_name(
                    shorten_name(
                        name_one, max_len=title_length, title_clip_pos=title_clip_pos
                    )
                )
                + ' [%s]' % aa_bp_unit,
                fontsize=label_size,
                fontweight='bold',
                labelpad=4,
            )
            P.ylabel(
                unicode_name(
                    shorten_name(
                        name_two, max_len=title_length, title_clip_pos=title_clip_pos
                    )
                )
                + ' [%s]' % aa_bp_unit,
                fontsize=label_size,
                fontweight='bold',
                labelpad=4,
            )
            P.tick_params(axis='both', which='major', labelsize=label_size * 0.9)

            # P.axis('scaled') # make images scaled by size ### optional update ###
            if not multi:
                if length_scaling:
                    ax.set_aspect(aspect='equal', adjustable='box', anchor='NW')
                P.xlim(0, len_one + 1)
                # xlimit = [0, len_one+1]
                if mirror_y_axis:
                    P.ylim(0, len_two + 1)  # rotate y axis (point upwards)
                else:
                    P.ylim(len_two + 1, 0)  # rotate y axis (point downwards)
            elif not length_scaling:
                P.xlim(0, len_one + 1)
                # xlimit = [0, len_one+1]
                if mirror_y_axis:
                    P.ylim(0, len_two + 1)  # rotate y axis (point upwards)
                else:
                    P.ylim(len_two + 1, 0)  # rotate y axis (point downwards)
            else:
                max_len = max(len_one, len_two)
                P.xlim(0, max_len + 1)
                # xlimit = [0, max_len+1]
                if mirror_y_axis:
                    P.ylim(0, max_len + 1)  # rotate y axis (point upwards)
                else:
                    P.ylim(max_len + 1, 0)  # rotate y axis (point downwards)

                # plot line deliminating shorter sequence
                if max_len != len_one:
                    ax.plot(
                        (len_one + 1, len_one + 1),
                        (0, len_two),
                        marker='',
                        linestyle='--',
                        color=scale_delim_col,
                        markerfacecolor='r',
                    )
                if max_len != len_two:
                    ax.plot(
                        (0, len_one),
                        (len_two + 1, len_two + 1),
                        marker='',
                        linestyle='--',
                        color=scale_delim_col,
                        markerfacecolor='r',
                    )

            # # use same tick labels for x and y axis
            # if P.xlim() == P.ylim():
            #     tick_locs, tick_labels = P.yticks()
            #     P.xticks(tick_locs)
            #     P.xlim(xlimit)

            # evtl. switch x axis position
            if x_label_pos_top:
                ax.xaxis.tick_top()
                ax.xaxis.set_label_position('top')
            P.setp(ax.get_xticklabels(), fontsize=label_size * 0.9)
            P.setp(ax.get_yticklabels(), fontsize=label_size * 0.9)

            # save figure and reinitiate if page is full
            if multi and counter == ncols * nrows:
                # finalize layout - margins & spacing between plots
                try:
                    P.tight_layout(h_pad=0.02, w_pad=0.02)
                except KeyError:
                    logging.warning(
                        'Attention - pylab.tight_layout failed! Please check sequence names and layout settings!'
                    )
                if x_label_pos_top:
                    P.subplots_adjust(
                        hspace=0.5, wspace=0.5, top=0.95
                    )  # space between rows - def 0.4
                else:
                    P.subplots_adjust(
                        hspace=0.5, wspace=0.5, bottom=0.05
                    )  # space between rows - def 0.4

                # name and create output files (names derived from SEQNAME)
                fig_name = '%s%s_wordsize%i%s-%.3d.%s' % (
                    prefix,
                    name_graph,
                    wordsize,
                    suffix,
                    page_counter,
                    filetype,
                )
                P.savefig(fig_name, bbox_inches='tight')
                P.close()
                P.cla()

                list_of_png_names.append(fig_name)

                counter = 0
                page_counter += 1

                P.figure(figsize=(figsize_x, figsize_y))

            # plotting separate figure files
            elif not multi:
                # finalize layout - margins & spacing between plots
                try:
                    P.tight_layout(h_pad=0.02, w_pad=0.02)
                except Exception:
                    logging.warning(
                        'Attention - pylab.tight_layout failed! Please check sequence names and layout settings!'
                    )
                if y_label_rotation == 'horizontal':
                    if x_label_pos_top:
                        P.subplots_adjust(
                            hspace=0.02, wspace=0.02, left=0.13, top=0.95
                        )  # space between rows - def 0.4
                    else:
                        P.subplots_adjust(
                            hspace=0.02, wspace=0.02, left=0.13, bottom=0.05
                        )  # space between rows - def 0.4
                else:
                    P.subplots_adjust(
                        hspace=0.02, wspace=0.02
                    )  # space between rows - def 0.4

                # name and create output files
                fig_name = '%s%s-%d_wordsize%i%s.%s' % (
                    prefix,
                    name_graph,
                    counter,
                    wordsize,
                    suffix,
                    filetype,
                )
                P.savefig(fig_name)
                P.close()
                P.cla()

                list_of_png_names.append(fig_name)
                P.figure()  # clear any prior graph

        if only_vs_first_seq:
            break

    # save figure
    if multi and counter >= 1:
        # finalize layout - margins & spacing between plots
        try:
            P.tight_layout(h_pad=0.02, w_pad=0.02)
        except Exception:
            logging.warning(
                'Attention - pylab.tight_layout failed! Please check sequence names and layout settings!'
            )
        if x_label_pos_top:
            P.subplots_adjust(
                hspace=0.5, wspace=0.5, top=0.95
            )  # space between rows - def 0.4
        else:
            P.subplots_adjust(
                hspace=0.5, wspace=0.5, bottom=0.05
            )  # space between rows - def 0.4

        # name and create output files (names derived from SEQNAME)
        fig_name = '%s%s_wordsize%i%s-%.3d.%s' % (
            prefix,
            name_graph,
            wordsize,
            suffix,
            page_counter,
            filetype,
        )
        P.savefig(fig_name, bbox_inches='tight')
        P.close()
        P.cla()

        list_of_png_names.append(fig_name)

    log_txt += '\n%d done' % seq_counter
    logging.info(log_txt)

    logging.debug(seq_text)

    return list_of_png_names


def polydotplot(
    input_fasta,
    wordsize=10,
    gff_files=None,
    alphabetic_sorting=False,
    convert_wobbles=False,
    filetype='png',
    gff_color_dict=None,
    input_user_matrix_file='',
    label_size=10,
    lcs_shading_interval_len=100,
    lcs_shading_num=5,
    lcs_shading_ori=0,
    lcs_shading_ref=0,
    lcs_shading=True,
    line_col_for='#000000',  # defalut black
    line_col_rev='#009243',  # default green
    line_width=1,
    max_N_percentage=10,
    mirror_y_axis=False,
    plot_size=10,
    prefix=None,
    representation=0,
    rotate_labels=False,
    spacing=0.04,
    substitution_count=0,
    title_clip_pos='B',
    title_length=float('Inf'),
    type_nuc=True,
    user_matrix_print=True,
    x_label_pos_top=True,
):
    """
    all-against-all dotplot
    derived from dotplot function

    lcs_shading_refs:
        0 color relative to maximum lcs observed in dataset [default]
        1 color by coverage of shorter sequence (e.g. lcs = 70% of seq1)
    lcs_shading_ori
        0 forward only
        1 reverse only
        2 both orientations (in opposite plot)
    """

    # Initialize default gff_color_dict if None
    if gff_color_dict is None:
        gff_color_dict = {'others': ('grey', 1, 0)}

    # Initialize default gff_files if None
    if gff_files is None:
        gff_files = []

    # read sequences
    seq_dict, sequences = read_seq(input_fasta)
    if seq_dict == {}:
        logging.warning('Failed to load sequences.')
        return []

    if type_nuc:
        aa_bp_unit = 'bp'
    else:
        aa_bp_unit = 'aa'

    if alphabetic_sorting:
        sequences = sorted(sequences)

    if len(sequences) == 0:
        text = '\n%s\n\nCreating %dx%d polydotplot image\n%s\n\n=>' % (
            50 * '=',
            len(sequences),
            len(sequences),
            30 * '-',
        )
        text += ' No sequences provided for polydotplot!\n\nTerminating polydotplot!'
        logging.info(text)
        return
    elif len(sequences) == 1:
        text = '\n\nCreating polydotplot for single sequence!'
        text += "\nRecommendation: Use selfdotplot via '--plotting_mode 0'!\n\n"
        logging.info(text)

    text = '\n%s\n\nCreating %dx%d polydotplot image\n%s\n\n=>' % (
        50 * '=',
        len(sequences),
        len(sequences),
        30 * '-',
    )
    text += ' ' + ' '.join(sequences) + '\n'
    logging.info(text)

    # read gff annotation data if provided for shading
    if gff_files is not None and gff_files != []:
        text = '\n%s\n\nReading %s GFF annotation files\n%s\n\n=> %s\n' % (
            50 * '=',
            len(gff_files),
            28 * '-',
            ', '.join(gff_files),
        )
        logging.info(text)
        if prefix is not None and prefix != '':
            legend_prefix = prefix + '-Polydotplot'
        else:
            legend_prefix = 'Polydotplot'
        feat_dict = read_gffs(
            gff_files,
            color_dict=gff_color_dict,
            type_nuc=type_nuc,
            prefix=legend_prefix,
            filetype=filetype,
        )

    if lcs_shading and not type_nuc:
        if lcs_shading_ori != 0:
            lcs_shading_ori = 0
            text = 'Protein shading does not support reverse complementary matching!\n'
            logging.info(text)

    # read custom shading matrix & match names of sequences to fasta
    if input_user_matrix_file != '' and input_user_matrix_file is not None:
        logging.info('Reading user matrix file: %s' % input_user_matrix_file)
        # lcs_shading_ori = 2
        custom_dict = read_matrix(input_user_matrix_file)
        if custom_dict != {}:
            custom_shading = True
            custom_similarity_dict = {}
            invalid_entries = []
            custom_max = 0
            custom_min = float('Inf')
            for key in list(custom_dict.keys()):
                number_key = []

                # convert number into float
                try:
                    value = float(custom_dict[key])
                    if '.' not in custom_dict[key]:
                        value = int(custom_dict[key])
                    custom_max = max(custom_max, value)
                    custom_min = min(custom_min, value)
                except Exception:
                    value = custom_dict[key]
                    if value == '':
                        value = None
                    invalid_entries.append(key)
                # match matrix names with sequence names
                for item in key:
                    if item in sequences:
                        number_key.append(sequences.index(item))
                    else:
                        number_key.append(-1)
                # dictionary with tuple of sorted sequence indices as key and number as value
                custom_similarity_dict[tuple(sorted(number_key))] = value
            if len(invalid_entries) != 0:
                text = (
                    'No valid number in custom similarity matrix for %d entries: \n\t'
                    % (len(invalid_entries))
                )
                for key in invalid_entries:
                    text += str(key) + ' - ' + str(custom_dict[key]) + '; '
                logging.info(text[:-2] + '\n')

        text = 'Custom user matrix given: min %.2f, max %.2f\n' % (
            custom_min,
            custom_max,
        )

        # artificially rounding intervals if likely identity/divergence percentages
        if 0 <= custom_min < 1 and 0 < custom_max <= 1:
            rounding_factor = 5
            multi_factor = 100
            text += (
                ' > artificially rounding custom shading intervals: old (%.2f, %.2f) - '
                % (custom_min, custom_max)
            )
            custom_min = max(
                0,
                (multi_factor * custom_min // rounding_factor)
                * (1.0 * rounding_factor / multi_factor),
            )
            custom_max = min(
                (multi_factor * custom_max // rounding_factor)
                * (1.0 * rounding_factor / multi_factor),
                1,
            )
            text += 'new (%.2f, >%2f)\n' % (custom_min, custom_max)

        elif 0 <= custom_min < 100 and 0 < custom_max <= 100:
            rounding_factor = 5
            text += (
                ' > artificially rounding custom shading intervals: old (%.2f, %.2f) - '
                % (custom_min, custom_max)
            )
            custom_min = max(0, (custom_min // rounding_factor) * rounding_factor)
            custom_max = min((custom_max // rounding_factor) * rounding_factor, 100)
            text += 'new (%d, >%d)\n' % (custom_min, custom_max)

        logging.info(text)

    else:
        custom_shading = False

    name_graph = 'Polydotplot'
    suffix = ''
    if convert_wobbles:
        suffix += '_wobbles'
    if substitution_count != 0:
        suffix += '_S%d' % substitution_count
    if custom_shading:
        suffix += '_matrix'
    if lcs_shading:
        suffix += '_%dshades_ref%d_ori%s' % (
            lcs_shading_num + 1,
            lcs_shading_ref,
            lcs_shading_ori,
        )
        if 'ref2' in suffix and type_nuc:
            suffix = suffix.replace('ref2', '%dbp' % lcs_shading_interval_len)
        elif 'ref2' in suffix:
            suffix = suffix.replace('ref2', '%daa' % lcs_shading_interval_len)

    # name and create output files (names derived from SEQNAME)
    if prefix:
        prefix = str(prefix) + '-'
    else:
        prefix = ''

    # preparations for background shading
    if lcs_shading or custom_shading:
        # create color range white to grey
        colors = create_color_list(lcs_shading_num + 1, color_map='Greys')
        colors_2 = create_color_list(lcs_shading_num + 1, color_map='OrRd')

        if custom_shading:
            text = 'Custom Matrix Colors: ' + ', '.join(colors_2)

    # write lcs lengths to file
    lcs_data_file = open(
        '%sPolydotplot_lcs_data_file%s.txt'
        % (prefix, suffix.replace('_scaled', '').replace('_collage', '')),
        'w',
    )
    lcs_data_file.write(
        '\t'.join(
            [
                '#title1',
                'title2',
                'len_seq1',
                'len_seq2',
                'len_lcs_for',
                '%_min_seq_len',
                'len_lcs_rev',
                '%_min_seq_len',
            ]
        )
        + '\n'
    )

    # compare sequences pairwise - save lcs and line information in dictionary for plotting
    data_dict = {}  # keys = tuple(idx, jdx), value = x1, y1, x2, y2 (line positions)
    lcs_dict = {}  # keys = tuple(idx, jdx), value = length of lcs: lcs_len or (lcs_for, lcs_rev)
    for_lcs_set = set()  # keep lengths to calculate max (excluding self comparisons)
    rev_lcs_set = set()  # keep lengths to calculate max (all)

    text = '\nTotal plot count:   %d' % (len(sequences) * (len(sequences)))
    text += '\nTotal calculations: %d' % (len(sequences) * (len(sequences) + 1) / 2)
    logging.info(text)

    logging.info(
        '\nCalculating shared regions and lengths of longest_common_substring...'
    )
    log_txt = '\nCalculating shared regions and lengths of longest_common_substring...'
    # determine  matches and length of lcs by comparing all sequence pairs

    seq_text = ''
    counter = 0
    for idx in range(len(sequences)):
        logging.debug('\n%d\t%s vs.' % ((counter + 1), sequences[idx]))
        seq_text += '\n%d\t%s vs.' % ((counter + 1), sequences[idx])
        rec_two = seq_dict[sequences[idx]]
        name_two = rec_two.id
        seq_two = rec_two.seq
        len_two = len(seq_two)

        for jdx in range(idx, len(sequences)):
            rec_one = seq_dict[sequences[jdx]]
            name_one = rec_one.id
            seq_one = rec_one.seq
            len_one = len(seq_one)

            counter += 1
            logging.debug(sequences[jdx])
            seq_text += ' ' + sequences[jdx]

            if len(sequences) < 5:
                log_txt += '\n\t%s (%d %s), %s (%d %s)' % (
                    name_one,
                    len_one,
                    aa_bp_unit,
                    name_two,
                    len_two,
                    aa_bp_unit,
                )
            else:
                if not counter % 25:
                    print(counter)
                    log_txt += str(counter)

            # Get positions of matches &  length of longest common substring based on match lengths
            if substitution_count != 0:
                # print "RE"
                x1, y1, x2, y2, lcs_for, lcs_rev = find_match_pos_regex(
                    seq_one,
                    seq_two,
                    wordsize,
                    substitution_count=substitution_count,
                    convert_wobbles=convert_wobbles,
                    max_N_percentage=max_N_percentage,
                    report_lcs=True,
                    type_nuc=type_nuc,
                )
            else:
                # print "DIAG"
                x1, y1, x2, y2, lcs_for, lcs_rev = find_match_pos_diag(
                    seq_one,
                    seq_two,
                    wordsize,
                    convert_wobbles=convert_wobbles,
                    max_N_percentage=max_N_percentage,
                    report_lcs=True,
                    type_nuc=type_nuc,
                )
            data_dict[(idx, jdx)] = x1[:], y1[:], x2[:], y2[:]
            lcs_dict[idx, jdx] = lcs_for, lcs_rev

            if idx != jdx:
                for_lcs_set.add(lcs_for)
            rev_lcs_set.add(lcs_rev)

            lcs_data_file.write(
                '\t'.join(
                    [
                        name_one,
                        name_two,
                        str(len_one),
                        str(len_two),
                        str(lcs_for),
                        str(round((lcs_for * 100.0 / min(len_one, len_two)), 3)),
                        str(lcs_rev),
                        str(round((lcs_rev * 100.0 / min(len_one, len_two)), 3)),
                    ]
                )
                + '\n'
            )

    log_txt += '\n' + str(len(sequences) * (len(sequences) + 1) / 2) + ' done'

    logging.info(log_txt)

    logging.debug('\n\nlcs_dict\n' + str(lcs_dict))
    if custom_shading:
        logging.debug('\ncustom_dict\n' + str(custom_dict))
        logging.debug('\ncustom_similarity_dict\n\n' + str(custom_similarity_dict))

    logging.info(seq_text + '\n')

    if lcs_shading_ref == 2:
        color_bins = []
        text = '\nLCS lengh bins: '
        for idx in range(lcs_shading_num):
            color_bins.append(lcs_shading_interval_len * (idx + 1))
            text += ' ' + str(lcs_shading_interval_len * (idx + 1))
        logging.info(text)

    # Calculate maximum lcs length
    if lcs_shading_ori == 0:  # forward only
        if len(for_lcs_set) != 0:
            max_lcs = max(for_lcs_set)
        else:
            max_lcs = None
    elif lcs_shading_ori == 1:  # reverse complement only
        if len(rev_lcs_set) != 0:
            max_lcs = max(rev_lcs_set)
        else:
            max_lcs = None
    else:  # both orientations
        if len(rev_lcs_set) != 0 and len(for_lcs_set) != 0:
            max_lcs = max(max(rev_lcs_set), max(for_lcs_set))
        elif len(rev_lcs_set) != 0:
            max_lcs = max(rev_lcs_set)
        elif len(for_lcs_set) != 0:
            max_lcs = max(for_lcs_set)
        else:
            max_lcs = None

    if max_lcs:
        text = 'Maximum LCS: %d %s' % (max_lcs, aa_bp_unit)
        logging.info(text)
    if custom_shading:
        text = 'Maximum custom value: %d\n' % custom_max
        logging.info(text)

    # count sequences
    ncols = len(sequences)
    nrows = len(sequences)

    # get sequence lengths to scale plot widths and heights accordingly
    size_ratios = []
    for item in sequences:
        size_ratios.append(len(seq_dict[item].seq))

    P.cla()  # clear any prior graph
    # use GridSpec to resize plots according to sequence length
    if mirror_y_axis:
        height_ratios = size_ratios[::-1]
    else:
        height_ratios = size_ratios[:]
    gs = gridspec.GridSpec(
        nrows, ncols, width_ratios=size_ratios, height_ratios=height_ratios
    )
    P.figure(figsize=(plot_size, plot_size))

    # for cartesian coordinate system with mirrored y-axis: plot x labels below plot
    if mirror_y_axis and representation == 1:
        x_label_pos_top = True
    elif mirror_y_axis or representation == 2:
        x_label_pos_top = False

    # print y labels on the right, if upper right triangle is displayed
    if (representation == 1 and not mirror_y_axis) or (
        representation == 2 and mirror_y_axis
    ):
        y_label_pos = 0  # last column
    else:  # left y label
        y_label_pos = 1  # first column

    # determine label orientations
    if len(sequences) > 5 or rotate_labels:
        x_label_rotation = 45
        y_label_rotation = 'horizontal'
        if x_label_pos_top:
            xhalign = 'left'
            xvalign = 'bottom'
        else:
            xhalign = 'right'
            xvalign = 'top'
        yhalign = 'right'
    else:
        x_label_rotation = 'horizontal'
        y_label_rotation = 'vertical'
        xvalign = 'center'
        xhalign = 'center'
        yhalign = 'center'
    yvalign = 'center'

    # check combination of shading parameters for triangular output
    if (
        representation != 0 and lcs_shading and custom_shading
    ):  # both directions in triangle
        logging.info(
            '\nAttention: For triangular output custom-shading and LCS shading cannot be combined!\n'
        )
    elif (
        representation != 0 and lcs_shading and lcs_shading_ori == 2
    ):  # both directions in triangle
        logging.info(
            '\nAttention: For triangular output LCS shading for both orientations is combined to max of both orientations!\n'
        )

    log_txt = '\nDrawing polydotplot...'

    # draw subplots
    if lcs_shading and custom_shading:
        lcs_text = (
            '\n'
            + '\t'.join(
                [
                    '#Seq1',
                    'Seq2',
                    'LCS for [%s]' % aa_bp_unit,
                    'LCS for [%s]' % aa_bp_unit,
                    'Custom matrix value',
                    'Matrix color index',
                    'LCS color index',
                ]
            )
            + '\n'
        )
    elif lcs_shading:
        lcs_text = (
            '\n'
            + '\t'.join(
                [
                    '#Seq1',
                    'Seq2',
                    'LCS for [%s]' % aa_bp_unit,
                    'LCS for [%s]' % aa_bp_unit,
                    'LCS color index for',
                    'LCS color index rev',
                ]
            )
            + '\n'
        )
    elif custom_shading:
        lcs_text = (
            '\n'
            + '\t'.join(
                [
                    '#Seq1',
                    'Seq2',
                    'Custom matrix value',
                    'Color index for',
                    'Color index rev',
                ]
            )
            + '\n'
        )

    seq_text = ''

    counter, seq_counter = 0, 0
    for idx in range(len(sequences)):
        logging.debug('\n%d\t%s vs.' % ((seq_counter + 1), sequences[idx]))
        seq_text += '\n%d\t%s vs.' % ((seq_counter + 1), sequences[idx])

        rec_two = seq_dict[sequences[idx]]
        len_two = len(rec_two.seq)
        name_two = rec_two.id

        for jdx in range(idx, len(sequences)):
            rec_one = seq_dict[sequences[jdx]]
            len_one = len(rec_one.seq)
            name_one = rec_one.id

            counter += 1
            seq_counter += 1

            logging.debug(sequences[jdx])
            seq_text += ' ' + sequences[jdx]

            if not seq_counter % 25:
                # print(seq_counter)
                log_txt += str(seq_counter)

            # optional shade background according to length of LCS and/or user matrix
            #########################################################################

            # get interval based on LCS
            background_colors = [None, None]
            if lcs_shading and (
                lcs_shading_ref == 1 or lcs_shading_ref == 2 or max_lcs is not None
            ):  # self plot max_lcs_for == None
                lcs_len = lcs_dict[(idx, jdx)]
                l1 = lcs_len[0]  # forward
                l2 = lcs_len[1]  # reverse complement

                lcs_shading_bool = True

                # calculate shading acc. to chosen option
                if lcs_shading_ref == 1:  # percentage of shorter sequence
                    color_idx0 = min(
                        len(colors) - 1, l1 * lcs_shading_num // min(len_one, len_two)
                    )
                    color_idx1 = min(
                        len(colors) - 1, l2 * lcs_shading_num // min(len_one, len_two)
                    )
                elif lcs_shading_ref == 2:  # by given interval size
                    color_idx0 = min(len(colors) - 1, l1 // lcs_shading_interval_len)
                    color_idx1 = min(len(colors) - 1, l2 // lcs_shading_interval_len)
                    if color_idx0 >= len(colors):
                        color_idx0 = len(colors)
                    if color_idx1 >= len(colors):
                        color_idx1 = len(colors)
                else:  # percentage of maximum lcs length
                    color_idx0 = min(len(colors) - 1, l1 * lcs_shading_num // max_lcs)
                    color_idx1 = min(len(colors) - 1, l2 * lcs_shading_num // max_lcs)
            else:
                lcs_shading_bool = False

            # get interval based on custom matrix
            if custom_shading:
                # matrix value
                try:
                    custom_value = custom_similarity_dict[(idx, jdx)]
                except Exception:
                    custom_value = ''

                # bottom left triangle = LCS forward/reverse or best of both
                if lcs_shading_bool:
                    if lcs_shading_ori == 0:  # forward
                        color_idx1 = color_idx0
                    elif lcs_shading_ori == 2:  # both directions
                        color_idx1 = max(color_idx0, color_idx1)

                # top right triangle = custom value (not colored if text matrix provided)
                if type(custom_value) is int or type(custom_value) is float:
                    color_idx0 = int(
                        (custom_value - custom_min)
                        * lcs_shading_num
                        // (custom_max - custom_min)
                    )
                # no color if string is proviced
                else:
                    color_idx0 = 0

            # use best LCS of both orientations for coloring triangle with two-ori-LCS
            if (
                representation != 0 and lcs_shading_ori == 2
            ):  # both directions in triangle
                color_idx0, color_idx1 = (
                    max(color_idx0, color_idx1),
                    max(color_idx0, color_idx1),
                )

            # set colors dependent on lcs dependent on orientation
            if lcs_shading_bool and not custom_shading:
                if idx != jdx:
                    if lcs_shading_ori == 0:
                        color_idx1 = color_idx0
                    elif lcs_shading_ori == 1:
                        color_idx0 = color_idx1
                    background_colors[0] = colors[color_idx0]
                    background_colors[1] = colors[color_idx1]
                # for selfcomparison, only color reverse complement
                elif lcs_shading_ori != 0 and not custom_shading:
                    background_colors[0] = colors[color_idx1]
            # set different colors for shading by LCS + user matrix
            elif lcs_shading_bool and custom_shading:
                # print colors, background_colors, color_idx0, color_idx1
                background_colors[0] = colors_2[color_idx0]
                background_colors[1] = colors[color_idx1]
            # set grey color range for user matrix if no LCS shading
            elif custom_shading:
                background_colors[0] = colors[color_idx0]
                background_colors[1] = colors[color_idx0]

            if custom_shading and lcs_shading_bool:
                lcs_text += (
                    '\t'.join(
                        [
                            name_one,
                            name_two,
                            str(lcs_len[0]),
                            str(lcs_len[1]),
                            str(custom_value),
                            str(color_idx0),
                            str(color_idx1),
                        ]
                    )
                    + '\n'
                )
            elif lcs_shading_bool:
                lcs_text += (
                    '\t'.join(
                        [
                            name_one,
                            name_two,
                            str(lcs_len[0]),
                            str(lcs_len[1]),
                            str(color_idx0),
                            str(color_idx1),
                        ]
                    )
                    + '\n'
                )
            elif custom_shading:
                lcs_text += (
                    '\t'.join(
                        [
                            name_one,
                            name_two,
                            str(custom_value),
                            str(color_idx0),
                            str(color_idx1),
                        ]
                    )
                    + '\n'
                )

            # calculate figure position in polyplot
            # diagonal (self-dotplots)
            if idx == jdx:
                if mirror_y_axis:
                    seq_num = sequences.index(name_one) + 1
                    counter1 = seq_num + len(sequences) * (len(sequences) - seq_num)
                    counter = counter + (counter - 1) // (nrows)
                else:
                    # skip positions below diagonal
                    counter1 = counter + (counter - 1) // (nrows)  # + row_pos
                    counter = counter1
                counters = [counter1]

            # draw both graphs at once (due to symmetry)
            else:
                if mirror_y_axis:
                    col_pos = sequences.index(name_two) + 1
                    row_pos = len(sequences) - (sequences.index(name_one) + 1)
                    counter1 = row_pos * ncols + col_pos
                    counter2 = (ncols - col_pos) * ncols + ncols - row_pos
                else:
                    counter1 = counter
                    col_pos = (counter - 1) % ncols
                    row_pos = (counter - 1) // (nrows)
                    counter2 = col_pos * ncols + row_pos + 1
                counters = [counter1, counter2]  # lower, upper

            if len(counters) == 2:
                seq_counter += 1
                if not seq_counter % 25:
                    # print(seq_counter)
                    log_txt += str(seq_counter)

            x_lists, y_lists, x_lists_rc, y_lists_rc = data_dict[(idx, jdx)]

            # plot diagram(s)
            for kdx in range(len(counters)):
                if (
                    representation == 0
                    or len(counters) == 1
                    or (representation == 1 and kdx == 0)
                    or (representation == 2 and kdx == 1)
                ):
                    fig_pos = counters[kdx]
                    # plotting subplot with matplotlib
                    ax = P.subplot(gs[fig_pos - 1])  # rows, columns, plotnumber

                    # shade annotated regions if gff file(s) provided
                    if idx == jdx and gff_files:
                        if name_one in list(feat_dict.keys()):
                            features = feat_dict[name_one]
                            if len_two != len_one:
                                logging.info(
                                    'Polydot GFF shading for diagonal fields - nequal length error!'
                                )
                                return
                            for item in features:
                                feat_type, start, stop = item
                                feat_color, strength, zoom = gff_color_dict[
                                    feat_type.lower()
                                ]
                                start = max(0, start - zoom - 0.5)
                                stop = min(len_one + 1, stop + zoom + 0.5)
                                width = stop - start
                                ax.add_patch(
                                    patches.Rectangle(
                                        (start, start),  # (x,y)
                                        width,
                                        width,  # width, height
                                        edgecolor=None,
                                        linewidth=line_width + zoom,
                                        fill=True,
                                        facecolor=feat_color,
                                        alpha=strength,
                                    )
                                )

                    # if custom matrix value printed into upper matrix triangle, skip data plotting
                    # text print in top triangle
                    if user_matrix_print and custom_shading and kdx == 0 and idx != jdx:
                        data_plotting = False
                    # dotplot in bottom triangle
                    else:
                        data_plotting = True

                    # mirror plot, if plotting below diagonal
                    if kdx == 0:
                        l1, l2 = len_one, len_two
                        n1, n2 = name_one, name_two
                        x1, y1 = x_lists, y_lists
                        x2, y2 = x_lists_rc, y_lists_rc
                    else:
                        l2, l1 = len_one, len_two
                        n2, n1 = name_one, name_two
                        x1, y1 = y_lists, x_lists
                        x2, y2 = y_lists_rc, x_lists_rc

                    if mirror_y_axis:
                        x1, y1, x2, y2 = y1, x1, y2, x2
                        n1, n2 = n2, n1

                    if data_plotting:
                        # collect lines
                        lines = []
                        color_list = []
                        for x_lines, y_lines, col in [
                            (x2, y2, line_col_rev),
                            (x1, y1, line_col_for),
                        ]:
                            # If color is not white, add lines to plot
                            if col != 'white':
                                for ldx in range(len(x_lines)):
                                    lines.append(
                                        [
                                            (x_lines[ldx][0], y_lines[ldx][0]),
                                            (x_lines[ldx][-1], y_lines[ldx][-1]),
                                        ]
                                    )
                                    color_list.append(col)
                        color_list = np.array(color_list)

                        # draw lines
                        lc = cllct.LineCollection(
                            lines, colors=color_list, linewidths=line_width
                        )
                        ax.add_collection(lc)

                    # plot value provided by customer instead of dotplot
                    else:
                        alignment = {
                            'horizontalalignment': 'center',
                            'verticalalignment': 'center',
                        }
                        # P.text(0.5, 0.5, custom_value, size='medium', transform=ax.transAxes, **alignment)
                        P.text(
                            0.5,
                            0.5,
                            custom_value,
                            size=label_size * 1.5,
                            transform=ax.transAxes,
                            **alignment,
                        )
                        # P.text(0.5, 0.5, custom_value, size=label_size*1.5, transform=ax.transAxes,
                        # horizontalalignment='center', verticalalignment='center', color="black")

                    if custom_shading:
                        # omit diagonal
                        if idx == jdx:
                            ax.set_facecolor('white')
                        # use white background for text fields (top right triangle only [kdx 0])
                        elif (
                            type(custom_value) is not int
                            and type(custom_value) is not float
                            and kdx == 0
                        ):
                            ax.set_facecolor('white')
                        else:
                            ax.set_facecolor(background_colors[kdx])
                    # set background color if lcs shading
                    elif lcs_shading_bool and background_colors[kdx]:
                        ax.set_facecolor(background_colors[kdx])

                    # set axis limits
                    # P.xlim(0, l1+1)
                    if mirror_y_axis:
                        P.xlim(0, l2 + 1)
                        P.ylim(0, l1 + 1)  # rotate y axis (point upwards)
                    else:
                        P.xlim(0, l1 + 1)
                        P.ylim(l2 + 1, 0)  # rotate y axis (point downwards)

                    ## axis labelling
                    ##################

                    # determine axis positions
                    if x_label_pos_top:
                        ax.xaxis.tick_top()
                        ax.xaxis.set_label_position('top')
                        x_label_bool = fig_pos <= ncols
                        x_tick_bool = fig_pos > ncols * (ncols - 1)
                    else:
                        x_label_bool = fig_pos > ncols * (ncols - 1)
                        x_tick_bool = fig_pos <= ncols

                    # settings for y labels on right side
                    if y_label_pos == 0:  # right label
                        ax.yaxis.tick_right()
                        ax.yaxis.set_label_position('right')
                        label_dist = 30
                    else:
                        label_dist = 8

                    # x axis labels dependent on plot position/number
                    if x_label_bool:  # x title and labels on top or bottom
                        P.xlabel(
                            unicode_name(
                                shorten_name(
                                    n1,
                                    max_len=title_length,
                                    title_clip_pos=title_clip_pos,
                                )
                            ),
                            fontsize=label_size,
                            rotation=x_label_rotation,
                            verticalalignment=xvalign,
                            horizontalalignment=xhalign,
                            fontweight='bold',
                            labelpad=8,
                        )  # axis naming
                        if x_label_rotation not in ['horizontal', 'vertical']:
                            P.setp(
                                ax.get_xticklabels(),
                                fontsize=label_size * 0.9,
                                rotation='vertical',
                            )
                        else:
                            P.setp(
                                ax.get_xticklabels(),
                                fontsize=label_size * 0.9,
                                rotation=x_label_rotation,
                            )
                    elif x_tick_bool and x_label_pos_top:  # x ticks on bottom row
                        ax.xaxis.tick_bottom()  # ticks without labels on bottom
                        P.setp(
                            ax.get_xticklabels(),
                            fontsize=label_size,
                            rotation=x_label_rotation,
                            visible=False,
                        )
                    elif x_tick_bool:  # x ticks on top row
                        ax.xaxis.tick_top()  # # ticks without labels on top
                        P.setp(
                            ax.get_xticklabels(),
                            fontsize=label_size,
                            rotation=x_label_rotation,
                            visible=False,
                        )  # inner diagrams without labelling
                    elif idx == jdx and representation != 0:
                        if not mirror_y_axis and representation == 1:  # upper
                            ax.xaxis.tick_bottom()
                        elif mirror_y_axis and representation == 2:  # lower
                            ax.xaxis.tick_top()
                        elif mirror_y_axis and representation == 1:  # upper
                            ax.xaxis.tick_bottom()
                        elif not mirror_y_axis and representation == 2:  # lower
                            ax.xaxis.tick_top()
                        P.setp(
                            ax.get_xticklabels(), visible=False
                        )  # inner diagrams without labelling
                    else:  # no x ticks on internal rows
                        ax.axes.get_xaxis().set_visible(False)

                    # y axis labels dependent on plot position/number
                    if fig_pos % ncols == y_label_pos or (
                        ncols == 1 and nrows == 1
                    ):  # y title and labels in 1st column
                        P.ylabel(
                            unicode_name(
                                shorten_name(
                                    n2,
                                    max_len=title_length,
                                    title_clip_pos=title_clip_pos,
                                )
                            ),
                            fontsize=label_size,
                            rotation=y_label_rotation,
                            verticalalignment=yvalign,
                            horizontalalignment=yhalign,
                            fontweight='bold',
                            labelpad=label_dist,
                        )
                        P.setp(
                            ax.get_yticklabels(), fontsize=label_size * 0.9
                        )  # axis naming
                    elif fig_pos % ncols == 0:  # y ticks in last column
                        ax.yaxis.tick_right()
                        P.setp(
                            ax.get_yticklabels(), visible=False
                        )  # inner diagrams without labelling
                    elif idx == jdx and representation != 0:
                        if not mirror_y_axis and representation == 1:  # upper
                            ax.yaxis.tick_left()
                        elif mirror_y_axis and representation == 2:  # lower
                            ax.yaxis.tick_left()
                        elif mirror_y_axis and representation == 1:  # upper
                            ax.yaxis.tick_right()
                        elif not mirror_y_axis and representation == 2:  # lower
                            ax.yaxis.tick_right()
                        P.setp(
                            ax.get_yticklabels(), visible=False
                        )  # inner diagrams without labelling
                    else:
                        ax.axes.get_yaxis().set_visible(False)

    log_txt += '\n%d done' % seq_counter
    logging.info(log_txt)

    try:
        logging.debug(lcs_text)
    except Exception:
        pass

    # finalize layout - margins & spacing between plots
    P.tick_params(axis='both', which='major', labelsize=label_size * 0.9)
    try:
        P.tight_layout(h_pad=0.02, w_pad=0.02)
    except Exception:
        logging.info(
            'Attention - pylab.tight_layout failed! Please check sequence names and layout settings!'
        )
    # gs.tight_layout(fig, h_pad=.02, w_pad=.02) # less overlapping tick labels, but also disturbingly large spacing
    if y_label_rotation == 'horizontal':
        if x_label_pos_top:
            P.subplots_adjust(
                hspace=spacing, wspace=spacing, left=0.13, top=0.87
            )  # space between rows - def 0.4
        else:
            P.subplots_adjust(
                hspace=spacing, wspace=spacing, left=0.13, bottom=0.13
            )  # space between rows - def 0.4
    else:
        P.subplots_adjust(
            hspace=spacing, wspace=spacing
        )  # space between rows - def 0.4

    # save figure and close instance
    fig_name = '%s%s_wordsize%i%s.%s' % (prefix, name_graph, wordsize, suffix, filetype)
    P.savefig(fig_name)
    P.close()
    P.cla()

    # create figure color legend
    if lcs_shading:
        if lcs_shading_ref == 1:  # percentage of shorter sequence
            legend_file_name = legend_figure(
                colors, lcs_shading_num, unit='%', filetype=filetype, prefix=prefix
            )
        elif lcs_shading_ref == 2:  # interval sizes
            legend_file_name = legend_figure(
                colors,
                lcs_shading_num,
                unit=aa_bp_unit,
                filetype=filetype,
                prefix=prefix,
                bins=color_bins,
            )
        else:  # relative of maximum lcs
            legend_file_name = legend_figure(
                colors,
                lcs_shading_num,
                unit=aa_bp_unit,
                filetype=filetype,
                prefix=prefix,
                max_len=max_lcs,
            )

    if custom_shading:
        custom_prefix = 'custom-matrix-' + prefix
        legend_file_name_custom = legend_figure(
            colors_2,
            lcs_shading_num,
            unit='%',
            filetype=filetype,
            prefix=custom_prefix,
            max_len=custom_max,
            min_len=custom_min,
        )

    if lcs_shading and custom_shading:
        return [fig_name, legend_file_name, legend_file_name_custom]
    elif lcs_shading:
        return [fig_name, legend_file_name]
    elif custom_shading:
        return [fig_name, legend_file_name_custom]
    else:
        return [fig_name]



import logging
import sys

from matplotlib import rc as mplrc
from matplotlib import rcParams
import pylab as P

from flexidot._version import __version__
from flexidot.plotting import pairdotplot, polydotplot, selfdotplot
from flexidot.utils.args import parse_args
from flexidot.utils.checks import check_kmer_length, print_summary
from flexidot.utils.file_handling import read_gff_color_config
from flexidot.utils.logs import init_logging

# Matplotlib settings

# Switch to non-interactive backend to avoid _tkinter.TclError on CentOs 7 servers see Github Issue #5
P.switch_backend('agg')

# Font settings
mplrc('pdf', fonttype=42, compression=0)

rcParams['font.family'] = 'sans-serif'
rcParams['font.sans-serif'] = [
    'Helvetica',
    'Verdana',
    'Tahoma',
    'DejaVu Sans',
    'Droid Sans Mono',
    'Sans',
    'Liberation',
    'Ubuntu',
    'Arial',
]

__citation__ = (
    'Please remember to cite FlexiDot as follows:\n\n'
    'Kathrin M Seibt, Thomas Schmidt, Tony Heitkam,\n'
    'FlexiDot: highly customizable, ambiguity-aware dotplots for visual sequence analyses,\n'
    'Bioinformatics, Volume 34, Issue 20, October 2018, Pages 3575–3577,\n'
    'https://doi.org/10.1093/bioinformatics/bty395'
)

###############################
#        Function Call        #
###############################


def main():
    # Parse command line arguments
    args = parse_args()

    # Set up logging
    init_logging(loglevel=args.loglevel, logfile=args.logfile)

    # Print summary of arguments
    print_summary(args)

    # Log version and command line arguments if debug is enabled
    logging.debug('FlexiDot version: %s' % __version__)
    logging.debug('  ${0}\n\n'.format(' '.join(sys.argv)))
    logging.debug('Command line arguments: %s' % args)

    # Check valid kmer length
    check_kmer_length(args.wordsize)

    # Set up variables
    alphabetic_sorting = args.sort
    convert_wobbles = args.wobble_conversion
    filetype = args.filetype
    gff = args.gff
    gff_color_config_file = args.gff_color_config
    input_user_matrix_file = args.user_matrix_file
    label_size = args.label_size
    lcs_shading_interval_len = args.lcs_shading_interval_len
    lcs_shading_num = args.lcs_shading_num
    lcs_shading_ori = args.lcs_shading_ori
    lcs_shading_ref = args.lcs_shading_ref
    lcs_shading = args.lcs_shading
    length_scaling = args.length_scaling
    max_N_percentage = args.max_n
    mirror_y_axis = args.mirror_y_axis
    modes = args.mode
    multi = args.collage
    ncols = args.n_col
    nrows = args.n_row
    only_vs_first_seq = args.only_vs_first_seq
    plot_size = args.plot_size
    prefix = f'{args.outdir}/{args.output_prefix}'
    norevcomp = args.norev
    seq_list = args.infiles
    spacing = args.spacing
    substitution_count = args.substitution_count
    title_clip_pos = (
        'B'  # Note: This was processed out of title_length in previouis versions
    )
    title_length = args.title_length
    user_matrix_print = args.user_matrix_print
    wordsize = args.wordsize
    line_col_rev = args.line_col_rev
    line_col_for = args.line_col_for
    line_width = args.line_width

    # Set True if nucleotide sequence
    if args.type_seq == 'nuc':
        type_nuc = True
    elif args.type_seq == 'aa':
        type_nuc = False

    # Set x label position
    if args.x_label_pos == 'top':
        x_label_pos_top = True
    elif args.x_label_pos == 'bottom':
        x_label_pos_top = False

    # Read gff color config file if provided
    if args.gff:
        if gff_color_config_file:
            logging.info(
                f'Reading GFF color configuration file: {gff_color_config_file}'
            )
            gff_feat_colors = read_gff_color_config(gff_color_config_file)
    else:
        gff_feat_colors = {}
        if gff_color_config_file:
            logging.warning(
                f'Provide GFF annotation files to use configuration file: {gff_color_config_file}'
            )

    # If color is set to white, reverse complementary matches are skipped
    if norevcomp:  # if norev is set
        line_col_rev = 'white'  # reverse matches not calculated

    if not type_nuc and not norevcomp:
        logging.warning('Reverse complement deactivated for proteins.')
        line_col_rev = 'white'  # reverse matches not calculated
    elif not type_nuc:
        line_col_rev = 'white'

    # Log plotting modes
    mode_text = []
    mode_names = {'0': 'self', '1': 'paired', '2': 'poly'}
    for item in modes:
        mode_text.append(str(item) + ': ' + mode_names[item])

    logging.info(f'Requested plotting modes: {", ".join(mode_text)}\n\n{50 * "="}')

    # Create dotplots
    ##########################################

    # Init empty list for image file names
    list_of_png_names = []

    # self dotplots
    if '0' in modes:
        logging.info('Calling selfdotplot')
        list_of_png_names = selfdotplot(
            seq_list,
            wordsize,
            alphabetic_sorting=alphabetic_sorting,
            convert_wobbles=convert_wobbles,
            filetype=args.filetype,
            gff_color_dict=gff_feat_colors,
            gff_files=gff,
            label_size=label_size,
            line_col_rev=line_col_rev,
            line_col_for=line_col_for,
            max_N_percentage=max_N_percentage,
            mirror_y_axis=mirror_y_axis,
            multi=multi,
            ncols=ncols,
            nrows=nrows,
            plot_size=plot_size,
            prefix=prefix,
            substitution_count=substitution_count,
            title_clip_pos=title_clip_pos,
            title_length=title_length,
            type_nuc=type_nuc,
            line_width=line_width,
        )
        if list_of_png_names:
            logging.info(
                f'\n-> Image file(s):\t{",\n\t\t\t".join(list_of_png_names)}\n\n{50 * "="}'
            )
        else:
            logging.warning(f'No image files were created!\n\n{50 * "="}\n')

    # paired dotplots
    if '1' in modes:
        if multi:
            logging.info('Calling pairdotplot with collage')
            list_of_png_names = pairdotplot(
                seq_list,
                wordsize,
                alphabetic_sorting=alphabetic_sorting,
                convert_wobbles=convert_wobbles,
                filetype=filetype,
                label_size=label_size,
                length_scaling=length_scaling,
                line_col_rev=line_col_rev,
                line_col_for=line_col_for,
                max_N_percentage=max_N_percentage,
                mirror_y_axis=mirror_y_axis,
                multi=multi,
                ncols=ncols,
                nrows=nrows,
                only_vs_first_seq=only_vs_first_seq,
                plot_size=plot_size,
                prefix=prefix,
                substitution_count=substitution_count,
                title_clip_pos=title_clip_pos,
                title_length=title_length,
                type_nuc=type_nuc,
                x_label_pos_top=x_label_pos_top,
                line_width=line_width,
            )
            # t1 = time_track(t1)
        else:
            if not length_scaling:
                logging.info(
                    'Pairwise dotplot with individual output files scaled by sequence length automatically.'
                )

            logging.info('Calling pairdotplot')
            list_of_png_names = pairdotplot(
                seq_list,
                wordsize,
                alphabetic_sorting=alphabetic_sorting,
                convert_wobbles=convert_wobbles,
                filetype=filetype,
                label_size=label_size,
                length_scaling=True,
                line_col_rev=line_col_rev,
                line_col_for=line_col_for,
                max_N_percentage=max_N_percentage,
                mirror_y_axis=mirror_y_axis,
                multi=multi,
                ncols=ncols,
                nrows=nrows,
                only_vs_first_seq=only_vs_first_seq,
                plot_size=plot_size,
                prefix=prefix,
                substitution_count=substitution_count,
                title_clip_pos=title_clip_pos,
                title_length=title_length,
                type_nuc=type_nuc,
                line_width=line_width,
            )
        if list_of_png_names:
            logging.info(
                f'\n-> Image file(s):\t{",\n\t\t\t".join(list_of_png_names)}\n\n{50 * "="}'
            )
        else:
            logging.warning(f'No image files were created!\n\n{50 * "="}\n')

    # all-against-all dotplot
    if '2' in modes:
        logging.info('Calling polydotplot')
        list_of_png_names = polydotplot(
            seq_list,
            wordsize,
            alphabetic_sorting=alphabetic_sorting,
            convert_wobbles=convert_wobbles,
            filetype=filetype,
            gff_color_dict=gff_feat_colors,
            gff_files=gff,
            input_user_matrix_file=input_user_matrix_file,
            label_size=label_size,
            line_col_rev=line_col_rev,
            line_col_for=line_col_for,
            lcs_shading_interval_len=lcs_shading_interval_len,
            lcs_shading_num=lcs_shading_num,
            lcs_shading_ori=lcs_shading_ori,
            lcs_shading_ref=lcs_shading_ref,
            lcs_shading=lcs_shading,
            max_N_percentage=max_N_percentage,
            mirror_y_axis=mirror_y_axis,
            plot_size=plot_size,
            prefix=prefix,
            representation=args.representation,
            spacing=spacing,
            substitution_count=substitution_count,
            title_clip_pos=title_clip_pos,
            title_length=title_length,
            type_nuc=type_nuc,
            user_matrix_print=user_matrix_print,
            line_width=line_width,
        )

        if list_of_png_names:
            logging.info(
                f'\n-> Image file(s):\t{",\n\t\t\t".join(list_of_png_names)}\n\n{50 * "="}'
            )
        else:
            logging.warning(f'No image files were created!\n\n{50 * "="}\n')

    logging.info(f'\nFinished! Thank you for using FlexiDot.\n\n{__citation__}')


######################
# FlexiDot Execution #
######################
if __name__ == '__main__':
    main()



from Bio.Seq import Seq

# Import the function to be tested
from flexidot.utils.matching import find_match_pos_diag


def test_basic_functionality_nucleotide():
    seq1 = Seq('ATCGATCG')
    seq2 = Seq('GATCGATC')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(seq1, seq2, wordsize, rc_option=False)
    assert len(x1) > 0
    assert len(y1) > 0
    assert len(x2) == 0
    assert len(y2) == 0


def test_basic_functionality_protein():
    seq1 = Seq('MKVLY')
    seq2 = Seq('KVLYM')
    wordsize = 3
    x1, y1, x2, y2 = find_match_pos_diag(
        seq1, seq2, wordsize, type_nuc=False, rc_option=False
    )
    assert len(x1) > 0
    assert len(y1) > 0
    assert len(x2) == 0
    assert len(y2) == 0


def test_reverse_complement_enabled():
    seq1 = Seq('ATCGATCG')
    seq2 = Seq('CGATCGAT')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(seq1, seq2, wordsize, rc_option=True)
    assert len(x1) > 0
    assert len(y1) > 0
    assert len(x2) > 0
    assert len(y2) > 0


def test_reverse_complement_disabled():
    seq1 = Seq('ATCGATCG')
    seq2 = Seq('CGATCGAT')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(seq1, seq2, wordsize, rc_option=False)
    assert len(x1) > 0
    assert len(y1) > 0
    assert len(x2) == 0
    assert len(y2) == 0


def test_ambiguous_residues_no_conversion():
    seq1 = Seq('ATCGNNNN')
    seq2 = Seq('NNNNGATC')
    wordsize = 4
    # When convert_wobbles is False, the function should skip any kmers with ambiguous residues
    x1, y1, x2, y2 = find_match_pos_diag(
        seq1, seq2, wordsize, convert_wobbles=False, max_N_percentage=60
    )
    assert len(x1) == 0
    assert len(y1) == 0
    assert len(x2) == 0
    assert len(y2) == 0


def test_ambiguous_residues_with_conversion():
    seq1 = Seq('ATCGNNNN')
    seq2 = Seq('NNNNGATC')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(
        seq1, seq2, wordsize, convert_wobbles=True, max_N_percentage=26
    )
    # One rev match ATCG -> NGAT
    assert len(x1) == 0
    assert len(y1) == 0
    assert len(x2) == 1
    assert len(y2) == 1


def test_report_lcs_nucleotide():
    seq1 = Seq('ATCGATCG')
    seq2 = Seq('GATCGATC')
    wordsize = 4
    x1, y1, x2, y2, lcs_for, lcs_rev = find_match_pos_diag(
        seq1, seq2, wordsize, report_lcs=True
    )
    assert lcs_for == 7
    assert lcs_rev == 7


def test_report_lcs_protein():
    seq1 = Seq('MKVLY')
    seq2 = Seq('KVLYM')
    wordsize = 3
    x1, y1, x2, y2, lcs_for, lcs_rev = find_match_pos_diag(
        seq1, seq2, wordsize, report_lcs=True, type_nuc=False
    )
    assert lcs_for > 0
    assert lcs_rev == 0


def test_short_sequences():
    seq1 = Seq('ATC')
    seq2 = Seq('GAT')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(seq1, seq2, wordsize)
    assert len(x1) == 0
    assert len(y1) == 0
    assert len(x2) == 0
    assert len(y2) == 0


def test_all_ambiguous_residues():
    seq1 = Seq('NNNN')
    seq2 = Seq('NNNN')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(seq1, seq2, wordsize)
    assert len(x1) == 0
    assert len(y1) == 0
    assert len(x2) == 0
    assert len(y2) == 0


def test_empty_sequences():
    seq1 = Seq('')
    seq2 = Seq('')
    wordsize = 4
    x1, y1, x2, y2 = find_match_pos_diag(seq1, seq2, wordsize)
    assert len(x1) == 0
    assert len(y1) == 0
    assert len(x2) == 0
    assert len(y2) == 0
